(*
 * asmgen.ml
 *)

open Tac_types
open Tacgen
open Tac_display
open Asm_types
open Asm_display
open Hex_print

(*
 * This source provides routines that turn a program specified by symbol
 * tables and Three Address Code (TAC) lists into assembly.
 *)

(*
 * sat_name_from_tac_ds
 *
 *      Provide a name for the assembly symbol alias table (SAT) based
 *      on the TAC dest/src provided.
 *
 * Output:  (name)
 *)
let sat_name_from_tac_ds ds =
        if ds.tds_type == Dvar then
               ("_var_" ^ ds.tds_name)
        else if ds.tds_type == Dtmp then
               ("_tmp_" ^ (string_of_int ds.tds_value))
        else if ds.tds_type == Darray then
               ("_arr_" ^ ds.tds_name)
        else
               let _ = print_tac_dest_src ds in
               raise (Failure ("Invalid type for sat name request"))

(* 
 * sate_from_tds
 *
 *      Provide a SAT element based on an input TAC dest/src.
 *
 * Output:  (sate)
 *)
let sate_from_tds ds sat =
        let name = sat_name_from_tac_ds ds in
        let sate = (if NameMap.mem name sat then
                (NameMap.find name sat)
                else
                (raise (Failure ("Name not found in sat: " ^
                        name)))) in
        (sate)

(*
 * sat_name_from_tac_symb
 *
 *      Provide a SAT name from a TAC symbol.
 *
 * Output:  (name)
 *)
let sat_name_from_tac_symb symb =
        if symb.symb_stype == Svar then
               ("_var_" ^ symb.symb_name)
        else if symb.symb_stype == Sarray then
               ("_arr_" ^ symb.symb_name)
        else
               raise (Failure ("sat_name_from_tac_symb requests name" ^
                        " for non var/array: " ^ symb.symb_name)) 

(*
 * exists_in_future
 *
 *      Determine whether or not a temporary value exists in the 
 *      future TAC list.  If it does not, then stack space allocated
 *      for the temporary variable may be reclaimed.
 *
 * Output:  (bool)
 *)
let rec exists_in_future tmpval ltac =
        if ltac == [] then false
        else let ctac = List.hd ltac in
        let destt = ctac.to_dest.tds_type in
        let src1t = ctac.to_src1.tds_type in
        let src2t = ctac.to_src2.tds_type in
        let destv = ctac.to_dest.tds_value in
        let src1v = ctac.to_src1.tds_value in
        let src2v = ctac.to_src2.tds_value in
        if src1t == Dtmp && src1v == tmpval then true
        else if src2t == Dtmp && src2v == tmpval then true
        else if destt == Dtmp && destv == tmpval then true
        else exists_in_future tmpval (List.tl ltac)



(* 
 * get_stack_space_for_temps
 *
 *      Determine the amount of stack space required for temporary
 *      values.  This is meant to be used with functions in order to
 *      determine stack management parameters when doing initialization.
 *
 *      NOTE:  We always set the maxinuse to 3 more than returned, this
 *      means that even if no temporaries are used beyond their current
 *      expression, stack space is available for all 3 dest/src if requested.
 *
 * Output: (maxinuse) 
 *)
let rec get_stack_space_for_temps ltac inuse maxinuse =
        if ltac == [] then (maxinuse + 3)
        else
                begin
                let tac = List.hd ltac in
                (* if the destination is a tmp, we incr inuse *)
                let dest_is_tmp = (tac.to_dest.tds_type == Dtmp) in
                let dest_future = 
                        (exists_in_future tac.to_dest.tds_value (List.tl ltac))
                in
                let inuse = if dest_is_tmp && dest_future then 
                        (inuse + 1) else inuse in
                (* dec if a source is not used in future *)
                let src1_is_tmp = (tac.to_src1.tds_type == Dtmp) in
                let src1_future = 
                        (exists_in_future tac.to_src1.tds_value (List.tl ltac))
                in
                let inuse = if src1_is_tmp && (src1_future == false) then
                        (inuse - 1) else inuse in
                let src2_is_tmp = (tac.to_src2.tds_type == Dtmp) in
                let src2_future = 
                        (exists_in_future tac.to_src2.tds_value (List.tl ltac))
                in
                let inuse = if src2_is_tmp && (src2_future == false) then
                        (inuse - 1) else inuse in
                (* if the current inuse is greater than the max, set it *)
                let maxinuse = (if inuse > maxinuse then inuse else maxinuse) in
                get_stack_space_for_temps (List.tl ltac) inuse maxinuse
                end
        

(**************************************************************************
 *
 *              METHODS TO SET UP SYMBOL ALIAS TABLE
 *
 *************************************************************************)

(* 
 * add_global_sat_symbols
 *
 *      Add global symbols to the current SAT symbol table.
 *
 * Output:  (sat, env) 
 *)
let rec add_global_sat_symbols lsymb sat env =
        if lsymb == [] then (sat, env)
        else let symb = List.hd lsymb in
        let lsymb = List.tl lsymb in
        if symb.symb_stype == Srcb || symb.symb_stype == Sfunc then
                (* functions and rcb's don't take up data space *)
                (add_global_sat_symbols lsymb sat env)
        else if symb.symb_stype == Svar then
                begin
                (* first we create the symbol name *)
                let sname = sat_name_from_tac_symb symb in
                let sat_symb = {
                        sat_name = sname;
                        sat_type = Sa_var;
                        sat_size = symb.symb_size;
                        sat_reg = Reg_none;
                        sat_loc = Sa_loc_data;
                        sat_soff = env.age_doff;
                        sat_nextuse = 0 } in
                (* this variable takes 4 bytes in the data segment *)
                (* increment the environment data offset by 4 *)
                let env = {
                        age_implabel = env.age_implabel;
                        age_soff = env.age_soff;
                        age_doff = env.age_doff + 4;
                        age_nlocals = env.age_nlocals } in
                let sat = 
                        (if NameMap.mem sname sat then
                                (raise (Failure ("sat name already appears" ^
                                        " within sat: " ^ sname)))
                        else (NameMap.add sname sat_symb sat)) in
               (add_global_sat_symbols lsymb sat env)
               end 
        else if symb.symb_stype == Sarray then
               begin
               (* first create the symbol name *)
               let sname = sat_name_from_tac_symb symb in
               let sat_symb = {
                       sat_name = sname;
                       sat_type = Sa_var;
                       sat_size = symb.symb_size;
                       sat_reg = Reg_none;
                       sat_loc = Sa_loc_data;
                       sat_soff = env.age_doff;
                       sat_nextuse = env.age_nlocals } in
               (* 
                * we reserve an extra 4-bytes of data space following the array
                * in order to allow unaligned accesses of the final bytes.
                *)
               let env = {
                       age_implabel = env.age_implabel;
                       age_soff = env.age_soff;
                       age_doff = env.age_doff + symb.symb_size + 4;
                       age_nlocals = env.age_nlocals } in
               let sat = 
                       (if NameMap.mem sname sat then
                                (raise (Failure ("sat name already appears" ^
                                        " within sat: " ^ sname)))
                        else (NameMap.add sname sat_symb sat)) in
               (add_global_sat_symbols lsymb sat env)
               end
        else
                raise (Failure ("unknown type passed to " ^
                        "add_global_sat_symbols"))


(*
 * Local symbols will make use of the stack almost exculsively.  The stack
 * within MOCV will look as follows:
 *
 *      +--------------+
 *      | Parameter1   | <- [ebp+16]
 *      +--------------+
 *      | Parameter2   | <- [ebp+12]
 *      +--------------+
 *      | Parameter3   | <- [ebp+8]
 *      +--------------+
 *      | Return Addr. | <- [ebp+4]
 *      +--------------+
 *      | Pushed EBP   | <- [ebp],  ebp = esp (following push ebp)
 *      +--------------+ 
 *      | Local1       | <- [ebp-4]
 *      +--------------+
 *      | Local2       | <- [ebp-8]
 *      +--------------+
 *      | Temp1        | <- [ebp-12]
 *      +--------------+
 *      | Temp2        | <- [ebp-16]
 *      +--------------+
 *           ... 
 *
 * During any function call, the code setting up the stack is as follows:
 *
 *      push    ebp
 *      mov     ebp,    esp
 *      sub     esp,    < size of locals + temps >
 *
 *      ; Then at the conclusion of the program, the stack manipulation is
 *      ; as follows
 *
 *      add     esp,    < size of locals + temps >
 *      pop     ebp
 *      ret
 *
 * This should maintain the stack properly regardless of the function calls.
 *)

(*
 * add_local_sat_symbols
 *
 *      Add local symbols to a SAT symbol table.
 *
 * Output:  (sat, env)
 *)
let rec add_local_sat_symbols lsymb sat env =
        if lsymb == [] then (sat, env)
        else let symb = List.hd lsymb in
        let lsymb = List.tl lsymb in
        if symb.symb_isparam == true then
                begin
                (* this is a parameter, it has a special location *)
                (*
                 * NOTE: In V8086 mode, addresses are alwasy 2 bytes, which
                 * means the parameters start at [ebp+6] and NOT [ebp+8]
                 *)
                let soff = 2 + 4 * symb.symb_paramnum in
                let sname = sat_name_from_tac_symb symb in
                let sat_symb = {
                        sat_name = sname;
                        sat_type = Sa_var;
                        sat_size = symb.symb_size;
                        sat_reg = Reg_none;
                        sat_loc = Sa_loc_stack;
                        sat_soff = soff;
                        sat_nextuse = 0 } in
                let sat = 
                        (if NameMap.mem sname sat then
                                (raise (Failure ("sat name already appears" ^
                                        " within sat: " ^ sname)))
                        else (NameMap.add sname sat_symb sat)) in
                (add_local_sat_symbols lsymb sat env)
                end
        else if symb.symb_stype == Srcb || symb.symb_stype == Sfunc then
                (* there shouldn't be any local function decls... *)
                raise (Failure ("function declared locally: " ^ symb.symb_name))
        else if symb.symb_stype == Sarray then
                begin
                (* all arrays are global - so we need to add this globally *)
                let g_lsymb = [symb] in
                let (sat, env) = add_global_sat_symbols g_lsymb sat env in
                (add_local_sat_symbols lsymb sat env)
                end
        else if symb.symb_stype == Svar then
                begin
                (* all local variables go on the stack *)
                let sname = sat_name_from_tac_symb symb in
                let sat_symb = {
                        sat_name = sname;
                        sat_type = Sa_var;
                        sat_size = symb.symb_size;
                        sat_reg = Reg_none;
                        sat_loc = Sa_loc_stack;
                        sat_soff = env.age_soff;
                        sat_nextuse = 0 } in
                (* stack is ebp relative, going down *)
                let sat = 
                        (if NameMap.mem sname sat then
                                (raise (Failure ("sat name already appears" ^
                                        " within sat: " ^ sname)))
                        else (NameMap.add sname sat_symb sat)) in
                (* the number of locals goes up by one, stack offset down by 4 *) 
                let env = {
                        age_implabel = env.age_implabel;
                        age_soff = env.age_soff - 4;
                        age_doff = env.age_doff;
                        age_nlocals = env.age_nlocals + 1 } in
                (add_local_sat_symbols lsymb sat env)
                end
        else 
                raise (Failure ("unknown symbol type in add_local_sat_symbols"))

(* 
 * add_tmp_sat_symbols
 *
 *      Add temporary symbols to a SAT symbol table.
 *
 * Output: (sat) 
 *)
let rec add_tmp_sat_symbols sat tt ltac =
        if ltac == [] then sat
        else
                begin
                let tac = List.hd ltac in
                (* if dest is tmp, add it to the table *)
                let dest_is_tmp = (tac.to_dest.tds_type == Dtmp) in
                let in_table = (if (dest_is_tmp == false) then false
                        else (NameMap.mem 
                        (sat_name_from_tac_ds tac.to_dest) sat)) in
                let (sat, tt) = 
                       (if (dest_is_tmp && (in_table == false)) then
                              begin
                              let nu = 0 in 
                              let n = sat_name_from_tac_ds tac.to_dest in
                              let t = Sa_tmp in
                              let r = Reg_none in
                              let l = Sa_loc_stack in
                              let o = ((List.hd tt.tt_free) + tt.tt_base) in
                              let sate = {
                                      sat_name = n;
                                      sat_type = t;
                                      sat_size = 4; (* TODO *)
                                      sat_reg = r;
                                      sat_loc = l;
                                      sat_soff = 0;
                                      sat_nextuse = nu } in
                              let sat = 
                                      (if NameMap.mem n sat then
                                              (raise (Failure ("name already " ^
                                                "exists in map: " ^ n)))
                                      else NameMap.add n sate sat) in
                              (* space allocated only if used in future *)
                              let ttfree = 
                                      (if nu != 0 then (List.tl tt.tt_free)
                                      else tt.tt_free) in
                              let tt = {
                                      tt_base = tt.tt_base;
                                      tt_free = ttfree } in
                              (sat, tt)
                              end
                       else (sat, tt)) in
                (* check the sources for something to free *)
                let update_tt_for_source sat tt ltac src =
                        let src_is_tmp = (src.tds_type == Dtmp) in
                        let tt = 
                                (if src_is_tmp then
                                        begin
                                        (* de-allocate if not used again *)
                                        let nu = 0 in
                                        let n = sat_name_from_tac_ds src in
                                        let tt = 
                                            (if nu == 0 then
                                                let sate = 
                                                        (if NameMap.mem n sat
                                                        then (NameMap.find n sat)
                                                        else 
                                                        raise (Failure ("Could not find name: " ^ 
                                                        n))) in
                                                let o = sate.sat_soff in
                                                let o = o - tt.tt_base in
                                                (* add it to free list *)
                                                let ttfree = o :: tt.tt_free in
                                                let tt = {
                                                        tt_base = tt.tt_base;
                                                        tt_free = ttfree } in
                                                (tt)
                                            else (tt)) in
                                        (tt)
                                        end
                                else (tt)) in
                        (tt) 
                in
                let tt = update_tt_for_source sat tt (List.tl ltac) tac.to_src1
                in
                let tt = update_tt_for_source sat tt (List.tl ltac) tac.to_src2
                in
                add_tmp_sat_symbols sat tt (List.tl ltac)
                end
                                                                

(* 
 * generate_local_init_assembly
 *
 *      If any of the local variables were declared initialized,
 *      provide assembly code to perform the initialization.
 *
 * Output:  (lasm) 
 *)
let rec generate_local_init_assembly sat lsymb lasm =
        if lsymb == [] then lasm
        else
                begin
                let symb = List.hd lsymb in
                let lsymb = List.tl lsymb in
                let sname = sat_name_from_tac_symb symb in
                let satsymb = 
                        (if NameMap.mem sname sat then NameMap.find sname sat
                        else raise (Failure ("generate_local_init_assembly:" ^
                                " symbol not found: " ^ sname))) in
                if symb.symb_init != 0 then
                        begin
                        let off = satsymb.sat_soff in
                        let off_str = 
                              (if satsymb.sat_soff > 0 then 
                                    ("+" ^ (string_of_int off))
                              else (string_of_int off)) in  
                        let asm = ("\tmov\tdword ptr ss:[ebp" ^ off_str ^
                                "],\t" ^ (string_of_int symb.symb_init)) in
                        let lasm = asm :: lasm in
                        (generate_local_init_assembly sat lsymb lasm)
                        end
                else (generate_local_init_assembly sat lsymb lasm)
                end


(*
 * update_rt
 *
 *      Update the Register Table based on information provided.
 *
 * Output:  (Register Table)
 *)
let update_rt reg symb rt free =
        if free then
                match reg with
                  Eax -> { rt_eax = ("invld", Rs_free);
                           rt_ebx = rt.rt_ebx;
                           rt_ecx = rt.rt_ecx;
                           rt_edx = rt.rt_edx;
                           rt_nextevict = Ebx }
                | Ebx -> { rt_eax = rt.rt_eax;
                           rt_ebx = ("invld", Rs_free);
                           rt_ecx = rt.rt_ecx;
                           rt_edx = rt.rt_edx;
                           rt_nextevict = Ecx }
                | Ecx -> { rt_eax = rt.rt_eax;
                           rt_ebx = rt.rt_ebx;
                           rt_ecx = ("invld", Rs_free);
                           rt_edx = rt.rt_edx;
                           rt_nextevict = Edx }
                | Edx -> { rt_eax = rt.rt_eax;
                           rt_ebx = rt.rt_ebx;
                           rt_ecx = rt.rt_ecx;
                           rt_edx = ("invld", Rs_free);
                           rt_nextevict = Eax }
                | Reg_none -> raise (Failure ("Request made to free Reg_non"))
        else
                match reg with
                  Eax -> { rt_eax = (symb.sat_name, Rs_used);
                           rt_ebx = rt.rt_ebx;
                           rt_ecx = rt.rt_ecx;
                           rt_edx = rt.rt_edx;
                           rt_nextevict = rt.rt_nextevict }
                | Ebx -> { rt_eax = rt.rt_eax;
                           rt_ebx = (symb.sat_name, Rs_used);
                           rt_ecx = rt.rt_ecx;
                           rt_edx = rt.rt_edx;
                           rt_nextevict = rt.rt_nextevict }
                | Ecx -> { rt_eax = rt.rt_eax;
                           rt_ebx = rt.rt_ebx;
                           rt_ecx = (symb.sat_name, Rs_used);
                           rt_edx = rt.rt_edx;
                           rt_nextevict = rt.rt_nextevict }
                | Edx -> { rt_eax = rt.rt_eax;
                           rt_ebx = rt.rt_ebx;
                           rt_ecx = rt.rt_ecx;
                           rt_edx = (symb.sat_name, Rs_used);
                           rt_nextevict = rt.rt_nextevict }
                | Reg_none -> raise (Failure ("Request made to free Reg_non"))

(*
 * string_of_reg
 *
 *      Based on a register input and a size, provide a string which
 *      represents the requested register.
 *
 * Output:  (register name)
 *)
let string_of_reg reg sz =
        let str = (if sz == 4 then "e" else "") in
        let str =
                (match reg with
                  Eax -> (str ^ "a")
                | Ebx -> (str ^ "b")
                | Ecx -> (str ^ "c")
                | Edx -> (str ^ "d")
                | Reg_none -> raise (Failure ("String_of_reg with Reg_none")))
        in
        let str = (if sz == 1 then (str ^ "l") else (str ^ "x")) in
        str

(* 
 * reg_to_mem
 *
 *      Generate assembly to put the contents of a register into 
 *      memory.
 *
 * Output:  (lasm) 
 *)
let reg_to_mem symb lasm =
        if symb.sat_loc == Sa_loc_stack then
                let imm = symb.sat_soff in
                let imm_str = (if imm > 0 then 
                        ("+" ^ (string_of_int imm))
                        else (string_of_int imm)) in
                let asm_str = ("\tmov\tss:[ebp" ^ imm_str ^
                        "],\t" ^ (string_of_reg symb.sat_reg symb.sat_size)) in
                let lasm = asm_str :: lasm in
                (lasm)
        else if symb.sat_loc == Sa_loc_data then
                let imm = symb.sat_soff in
                let imm_str = (if imm < 0 then 
                       (raise (Failure ("Data segment immediate " ^
                        "is less than zero!  " ^ symb.sat_name)))
                        else (string_of_int symb.sat_soff)) in
                let asm_str = ("\tmov\tds:[" ^ imm_str ^
                        "],\t" ^ (string_of_reg symb.sat_reg symb.sat_size)) in
                let lasm = asm_str :: lasm in
                (lasm)
        else
                raise (Failure ("Location of reg_to_mem " ^
                        "is invalid: " ^ symb.sat_name))
                 

(* 
 * evict_reg
 *
 *      Evict a particular register, moving the contents from the 
 *      register back into memory if necessary.
 *
 * Output:  (sat, rt, tt, lasm)
 *)
let evict_reg reg sat rt tt lasm =
        (* see if the register is already free - if so, return *)
        let reg_status = 
                match reg with
                  Eax -> snd rt.rt_eax 
                | Ebx -> snd rt.rt_ebx
                | Ecx -> snd rt.rt_ecx
                | Edx -> snd rt.rt_edx
                | Reg_none -> raise (Failure ("Reg_none passed into evict_reg"))
        in
        if reg_status == Rs_free then (sat, rt, tt, lasm)
        (* get the symbol name for the current reg entry *)
        else let pname = 
                match reg with
                  Eax -> fst rt.rt_eax
                | Ebx -> fst rt.rt_ebx
                | Ecx -> fst rt.rt_ecx
                | Edx -> fst rt.rt_edx
                | Reg_none -> raise (Failure ("Reg_none passed into evict_reg"))
        in
        if ((String.compare pname "invld") == 0) then
                (* this is invalid - must be immediate, just evict *)
                let rt = update_rt reg invld_sat_symb rt true in
                (sat, rt, tt, lasm)
        else
        let psymb = 
                (if NameMap.mem pname sat then (NameMap.find pname sat)
                else raise (Failure ("Could not find psymb: " ^ pname))) in
        (* Make sure reg in symbol matches the passed in reg *)
        let _ = (if psymb.sat_reg != reg then
                        let _ = debug_print_rt rt in
                        let _ = print_sat sat in
                        (raise (Failure ("Symbol reg (" ^ (string_of_reg
                        psymb.sat_reg 4) ^ ") does not equal reg used to obtain it! (" ^ 
                        (string_of_reg reg 4) ^ ") - " ^ psymb.sat_name)))
                else 0) in
        (* On eviction, we need to: 
         *      1.  Move symbol from memory to its proper location
         *      2.  Set symbols stack offset properly 
         *      3.  Set symbols register to invalid
         *      4.  Clear the Register Table setting
         *)
        if (psymb.sat_type == Sa_var ||
            psymb.sat_type == Sa_tmp) then
                (* Place variable back in memory *)
                let lasm = reg_to_mem psymb lasm in
                let nsymb = { sat_name = psymb.sat_name;
                              sat_type = psymb.sat_type;
                              sat_size = psymb.sat_size;
                              sat_reg = Reg_none;
                              sat_loc = psymb.sat_loc;
                              sat_soff = psymb.sat_soff;
                              sat_nextuse = 0 } in
                (* update the SAT *)
                let sat = NameMap.remove pname sat in
                let sat = NameMap.add pname nsymb sat in
                (* Update the RT to show the eviction *)
                let rt = update_rt reg invld_sat_symb rt true in
                (sat, rt, tt, lasm) 
        else
                (* invld - we should never see this *)
                raise (Failure ("eviction of an invld resource"))
     
(*
 * evict_all_reg
 *
 *      Evict all registers.
 *
 * Output:  (sat, rt, tt, lasm)
 *)
let evict_all_reg sat rt tt lasm =
        let (sat, rt, tt, lasm) = evict_reg Eax sat rt tt lasm in
        let (sat, rt, tt, lasm) = evict_reg Ebx sat rt tt lasm in
        let (sat, rt, tt, lasm) = evict_reg Ecx sat rt tt lasm in
        let (sat, rt, tt, lasm) = evict_reg Edx sat rt tt lasm in
        (sat, rt, tt, lasm)



(* 
 * get_reg_resource
 * 
 *      Request a register resource.  Request can either be for a 
 *      particular register, or may be just a generic request for
 *      any register.
 *
 * Output:  (sat, rt, tt, lasm, reg)
 *)
let get_reg_resource tac_symb tac_list sat rt tt lasm reg_req =
        let name = 
                (if (tac_symb.tds_type == Dtmp || 
                     tac_symb.tds_type == Dvar) then 
                        (sat_name_from_tac_ds tac_symb)
                else 
                        let _ = print_tac_dest_src tac_symb in
                        raise (Failure ("non tmp/var requesting reg"))) in
        let symb = 
                (if NameMap.mem name sat then (NameMap.find name sat)
                else 
                        begin
                        let _ = print_endline "*** FAILURE:  COULD NOT FIND SYMB:" in
                        let _ = print_sat sat in
                        let _ = debug_print_rt rt in
                        let _ = print_tac_dest_src tac_symb in
                        let _ = print_endline "\tREMAINING TAC LIST" in
                        let _ = print_tac_list tac_list in
                        raise (Failure ("get_reg_resource: could not find symb: " ^
                                name))
                        end) in
        (* if there is a specific register request, we must oblige *)
        if reg_req != Reg_none then
                (* a specific register is requested *)
                (* if this symb is already in that reg, just return *)
                if symb.sat_reg == reg_req then
                       (* we already own the register we want, just return *)
                       (sat, rt, tt, lasm, reg_req)
                else
                (* now evict the register we need *) 
                let (sat, rt, tt, lasm) = evict_reg reg_req sat rt tt lasm in
                let reg = reg_req in

                (* if this symb has a register, evict it first *)
                let (sat, rt, tt, lasm) = (if symb.sat_reg != Reg_none then
                       (* we have a register already, evict it *)
                       (evict_reg symb.sat_reg sat rt tt lasm)
                       else (sat, rt, tt, lasm)) in
                let symb = {
                        sat_name = symb.sat_name;
                        sat_type = symb.sat_type;
                        sat_size = symb.sat_size;
                        sat_reg = reg;
                        sat_loc = symb.sat_loc;
                        sat_soff = symb.sat_soff;
                        sat_nextuse = 0
                        } in
                (* remove the previous SAT entry, add the updated one *)
                let sat = NameMap.remove name sat in
                let sat = NameMap.add name symb sat in
                (* update rt to reflect register ownership *)
                let rt = update_rt reg symb rt false in 
                (sat, rt, tt, lasm, reg)
        (* if it's already using a register, then great *)
        else if symb.sat_reg != Reg_none then (sat, rt, tt, lasm, symb.sat_reg)
        (* next, try to allocate a free register *)
        else let reg = 
                (if snd rt.rt_eax == Rs_free then Eax
                else if snd rt.rt_ebx == Rs_free then Ebx
                else if snd rt.rt_ecx == Rs_free then Ecx
                else if snd rt.rt_edx == Rs_free then Edx
                else Reg_none) in
        if reg != Reg_none then
                (* there is a free register, use it *)
                let symb = {
                        sat_name = symb.sat_name;
                        sat_type = symb.sat_type;
                        sat_size = symb.sat_size;
                        sat_reg = reg;
                        sat_loc = symb.sat_loc;
                        sat_soff = symb.sat_soff;
                        sat_nextuse = 0
                        } in
                (* update the rt *)
                let rt = update_rt reg symb rt false in
                (* remove the previous SAT entry, add the updated one *)
                let sat = NameMap.remove name sat in
                let sat = NameMap.add name symb sat in 
                (sat, rt, tt, lasm, reg)
        else
                (* there are no free registers, we need to take one *)
                let reg = rt.rt_nextevict in
                (* first evict the current contents of register *)
                let (sat, rt, tt, lasm) = evict_reg reg sat rt tt lasm in
                (* create updated symbol with register information *)
                let symb = {
                        sat_name = symb.sat_name;
                        sat_type = symb.sat_type;
                        sat_size = symb.sat_size;
                        sat_reg = reg;
                        sat_loc = symb.sat_loc;
                        sat_soff = symb.sat_soff;
                        sat_nextuse = 0
                        } in
                (* update the rt *)
                let rt = update_rt reg symb rt false in
                (* update the symbol alias table *)
                let sat = NameMap.remove name sat in
                let sat = NameMap.add name symb sat in
                (sat, rt, tt, lasm, reg)

(* 
 * get_src_string
 *
 *      Get a string that represents a TAC dest/src type in assembly.
 *
 * Output: (source string) 
 *)
let get_src_string src sat reg_size =
        if src.tds_type == Darray then 
                begin
                (* the correct DS offset is in the sat sat_soff *)
                let src_n = sat_name_from_tac_ds src in
                let sate = 
                        (if NameMap.mem src_n sat then NameMap.find src_n sat
                        else raise (Failure ("Name not found in sat: " ^
                                src_n))) in
                (string_of_int sate.sat_soff)
                end
        else if src.tds_type == Dimm then (string_of_int src.tds_value)
        else
        let src_n = sat_name_from_tac_ds src in
        let sate = 
                (if NameMap.mem src_n sat then NameMap.find src_n sat
                else raise (Failure ("Name not found in sat: " ^ src_n)))
        in
        let sz = src.tds_size in
        (*
         * Three cases:
         *      1.  It's already in a register (just use it)
         *      2.  It's a tmp and it's on the stack
         *      3.  It's a var and it's on the stack
         *)
        if sate.sat_reg != Reg_none then
                (* it's already in a reg - perfect *)
                let sz = (if reg_size != 0 then reg_size
                        else sz) in
                (string_of_reg sate.sat_reg sz)
        else
                let ptr = (if sz == 1 then "byte ptr " 
                        else if sz == 2 then "word ptr "
                        else if sz == 4 then "dword ptr "
                        else raise (Failure ("get_src_string: sz = " ^
                                        (string_of_int sz)))) in
                let imm = sate.sat_soff in
                let imm_str =
                        (if imm > 0 then ("+" ^ (string_of_int imm))
                        else (string_of_int imm)) in
                let imm_str_data = string_of_int imm in
                (* it's not in a register... *)
                if sate.sat_type == Sa_tmp || sate.sat_type == Sa_var then
                        if sate.sat_loc == Sa_loc_stack then
                                (* it's on the stack *)
                                (ptr ^ "ss:[ebp" ^ imm_str ^ "]")
                        else if sate.sat_loc == Sa_loc_data then
                                (* it's in the data segment *)
                                (ptr ^ "ds:[" ^ imm_str_data ^ "]")
                        else
                                raise (Failure ("unknown memory location!"))
                else raise (Failure ("non tmp/var asks for src string " ^
                                        sate.sat_name)) 

(* 
 * place_in_reg
 *
 *      Place a TAC dest/src type into a register.  This is required
 *      for instructions that must be executed on at least one 
 *      register operand.
 *
 * Output:  (reg, sat, rt, tt, lasm) 
 *)
let place_in_reg tds ltac sat rt tt lasm =
        match tds.tds_type with
           Dvar | Dtmp -> let n = sat_name_from_tac_ds tds in
                let symb = (if NameMap.mem n sat then (NameMap.find n sat)
                        else (raise (Failure ("Name not found in sat " ^ n))))
                in
                if symb.sat_reg != Reg_none then
                        (symb.sat_reg, sat, rt, tt, lasm)
                else
                let src_str = get_src_string tds sat 0 in
                let (sat, rt, tt, lasm, reg) = 
                        get_reg_resource tds ltac sat rt tt lasm Reg_none in
                (* get the string for the first source operand *)
                let dest_str = string_of_reg reg tds.tds_size in
                (* perform the assignment *)
                let asm = ("\tmov\t" ^ dest_str ^ ",\t" ^ src_str) in
                let lasm = asm :: lasm in
                (reg, sat, rt, tt, lasm)
         | Dimm -> (* an immediate, we must move it into a reg *)
                let reg = rt.rt_nextevict in
                let (sat, rt, tt, lasm) =
                       evict_reg reg sat rt tt lasm in
                let asm = ("\tmov\t" ^ (string_of_reg reg tds.tds_size) ^
                        ",\t" ^ (string_of_int tds.tds_value)) in
                let lasm = asm :: lasm in
                (reg, sat, rt, tt, lasm)
         | Darray -> raise (Failure ("source input is an array"))

(* 
 * string_from_tacop
 *
 *      Get a string that represents a TAC operation in assembly.  Note
 *      that not all TAC operations have valid string representations
 *      within assembly.
 *
 * Output:  (string) 
 *)
let string_from_tacop op =
       match op with
          Tadd -> "add"
        | Tsub -> "sub"
        | Tmul -> "mul"
        | Tdiv -> "div"
        | Trdiv -> "div"
        | Txor -> "xor"
        | Tor -> "or"
        | Tand -> "and"
        | Teq -> "je"
        | Tne -> "jne"
        | Tlt -> "jl"
        | Tgt -> "jg"
        | Tle -> "jle"
        | Tge -> "jge"
        | Tuminus -> "neg"
        | Tnot -> "not"
        | Tjmp -> "jmp"
        | Tjmpe -> "je"
        | Tjmpne -> "jne"
        | Tjmpgt -> "jg"
        | Tjmplt -> "jl"
        | Tjmpge -> "jge"
        | Tjmple -> "jle"
        | Tlabel -> "L"
        | Tarray -> "invld"
        | Tasn -> "mov"
        | Taasn -> "invld"
        | Tcmp -> "cmp"
        | Tret -> "ret"
        | Tparam -> "invld"
        | Tcall -> "call"
        | Tevictall -> "invld"
        | Tespadd -> "add"
        | Tevicteax -> "invld"
        | Tevictebx -> "invld"
        | Tevictecx -> "invld"
        | Tevictedx -> "invld"


(* 
 * binary_op
 *
 *      Provide assembly for a binary TAC operation.
 *
 * The supported ops are:
 *
 *      add,    sub,    xor,    or,     and
 *
 *
 * Intended assembly for binary operations
 *
 *      mov     dest,   src1
 *      op      dest,   src2
 *
 * Output: (lasm, sat, rt, tt, env) 
 *)
let binary_op ltac lasm sat rt tt env =
        let tac_el = List.hd ltac in
        let op = tac_el.to_op in
        let dest = tac_el.to_dest in
        let src1 = tac_el.to_src1 in
        let src2 = tac_el.to_src2 in
        (* first get the register for the temporary *) 
        let (sat, rt, tt, lasm, reg) =
               get_reg_resource dest ltac sat rt tt lasm Reg_none in
        (* get the string for the first source operand *)
        let src1_str = get_src_string src1 sat 0 in
        (* get the string for the second source operand *)
        let src2_str = get_src_string src2 sat 0 in
        (* get the string for the destination register *)
        let dest_str = string_of_reg reg dest.tds_size in
        (* get the string for the operation *)
        let op_str = string_from_tacop op in
        (* finally, create the assembly string and add it to the list *)
        let asm = ("\tmov\t" ^ dest_str ^ ",\t" ^ src1_str) in
        let lasm = asm :: lasm in
        let asm = ("\t" ^ op_str ^ "\t" ^ dest_str ^ ",\t" ^ src2_str) in
        let lasm = asm :: lasm in
        (lasm, sat, rt, tt, env)


(* 
 * comparison_op
 *
 *      Provide assembly for a TAC comparison operation.
 *
 * The supported comparison operations are
 *
 *      eq,     ne,     lt,     gt,     le,     ge
 *
 * The intended assembly for these operations is as follows
 *
 *              cmp     src1,   src2
 *              j_op    IL#1
 *              mov     dest,   0
 *              jmp     IL#2
 *      IL#1:   mov     dest,   1
 *      IL#2:
 *
 * NOTE:  The reason for this complexity is that some users may wish
 * to use the result of a comparison in an expression.  The x86 architecture
 * does comparisons through flags, so in order to move the result into a 
 * temporary register, we need to add this additional complexity.
 *
 * Output:  (lasm, sat, rt, tt, env) 
 *)
let comparison_op ltac lasm sat rt tt env =
        let tac_el = List.hd ltac in
        let op = tac_el.to_op in
        let dest = tac_el.to_dest in
        let src1 = tac_el.to_src1 in
        let src2 = tac_el.to_src2 in
        (* first get the register for the temporary *) 
        let (sat, rt, tt, lasm, reg) =
               get_reg_resource dest ltac sat rt tt lasm Reg_none in
        (* get the string for the first source operand *)
        (* if the source is not in a reg, we need to move it there *)
        (* if the source is an immediate, we need to move it to a reg *)
        let (regs1, sat, rt, tt, lasm) = 
                place_in_reg src1 ltac sat rt tt lasm in
        let _ = (if regs1 != reg then 1
                else raise (Failure("comparison_op: tmp reg and src1 reg " ^
                        "are the same!"))) in
        let src1_str = string_of_reg regs1 src1.tds_size in
        (* get the string for the second source operand *)
        let src2_str = get_src_string src2 sat 0 in
        (* get the string for the destination register *)
        let dest_str = string_of_reg reg dest.tds_size in
        (* get the string for the operation *)
        let jcmp_str = string_from_tacop op in
        (* create the implicit label numbers *)
        let il1 = env.age_implabel in
        let env = { 
                age_implabel = env.age_implabel + 1; 
                age_soff = env.age_soff;
                age_doff = env.age_doff;
                age_nlocals = env.age_nlocals } in
        let il2 = env.age_implabel in
        let env = { 
                age_implabel = env.age_implabel + 1;
                age_soff = env.age_soff;
                age_doff = env.age_doff;
                age_nlocals = env.age_nlocals } in 
        (* finally, create the assembly string and add it to the list *)
        let asm = ("\tcmp\t" ^ src1_str ^ ",\t" ^ src2_str) in
        let lasm = asm :: lasm in
        let asm = ("\t" ^ jcmp_str ^ "\t" ^ "IL" ^ (string_of_int il1)) in
        let lasm = asm :: lasm in
        let asm = ("\tmov\t" ^ dest_str ^ ",\t0") in
        let lasm = asm :: lasm in
        let asm = ("\tjmp\tIL" ^ (string_of_int il2)) in
        let lasm = asm :: lasm in
        let asm = ("IL" ^ (string_of_int il1) ^ ":\tmov\t" ^ dest_str ^ ",\t1")
        in
        let lasm = asm :: lasm in
        let asm = ("IL" ^ (string_of_int il2) ^ ":") in
        let lasm = asm :: lasm in
        (lasm, sat, rt, tt, env)


(* 
 * jmp_op
 *
 *      Provide assembly for a jump operation.
 *
 * Jump operations supported are as follows:
 *
 *      jmp,    je,     jne,    jg,     jl      jge,    jle
 *
 * The intended assembly for this is as follows:
 *
 *      j_op    dest_label
 * 
 * Output:  (lasm, sat, rt, tt, env) *)
let jmp_op ltac lasm sat rt tt env =
        let tac_el = List.hd ltac in
        let op = tac_el.to_op in
        let dest = tac_el.to_dest in
        (* get the name of the label we are jumping too *)
        let l_n = ("L" ^ (string_of_int (dest.tds_value))) in
        (* get the string of the jmp operation being performed *)
        let op_str = string_from_tacop op in
        (* generate the assembly *)
        let asm = ("\t" ^ op_str ^ "\t" ^ l_n) in
        let lasm = asm :: lasm in
        (lasm, sat, rt, tt, env)


(* 
 * unary_op
 *
 *      Provide assembly for a unary operation.
 *
 * Supported unary operations are as follows:
 *
 *      not,    neg
 *
 * The intended assembly is:
 *
 *      mov     dest,   src
 *      op      dest
 * 
 * Output: (lasm, sat, rt, tt, env) 
 *)
let unary_op ltac lasm sat rt tt env =
        let tac_el = List.hd ltac in
        let op = tac_el.to_op in
        let dest = tac_el.to_dest in
        let src = tac_el.to_src1 in
        (* first get the register for the temporary *) 
        let (sat, rt, tt, lasm, reg) =
               get_reg_resource dest ltac sat rt tt lasm Reg_none in
        (* get the string for the first source operand *)
        let src_str = get_src_string src sat 0 in
        (* get the string for the destination register *)
        let dest_str = string_of_reg reg dest.tds_size in
        (* get the string for the operation *)
        let op_str = string_from_tacop op in
        (* generate the assembly *)
        let asm = ("\tmov\t" ^ dest_str ^ ",\t" ^ src_str) in
        let lasm = asm :: lasm in
        let asm = ("\t" ^ op_str ^ "\t" ^ dest_str) in
        let lasm = asm :: lasm in
        (lasm, sat, rt, tt, env)

(*
 * div_op
 *
 *      Provide assembly for a divide operation.
 *
 * Supported divide operations are as follows:
 *
 *      div,    div (remainder)
 *
 *
 * The tac op is as follows:  < div / rdiv, dest, src1, src2 >  
 * Intended assembly is as follows:
 * 
 *      8-bit division:  AX / src = AL (rem AH)
 *      16-bit division:  DX:AX / src = AX (rem DX)
 *      32-bit division:  EDX:EAX / src = EAX (rem EDX)
 *
 *      < evict EAX >
 *      < evict EDX >
 *      mov     RAX,    src1
 *      div     src2
 *      mov     dest,   REGA ; if div
 *      mov     dest,   REGD ; if rdiv
 *
 * Output: (lasm, sat, rt, tt, env) 
 *)
let div_op ltac lasm sat rt tt env =
        let tac_el = List.hd ltac in
        let op = tac_el.to_op in
        let dest = tac_el.to_dest in
        let src1 = tac_el.to_src1 in
        let src2 = tac_el.to_src2 in

        (* evict the contents of both EAX and EDX *)
        let (sat, rt, tt, lasm) = evict_reg Eax sat rt tt lasm in
        let (sat, rt, tt, lasm) = evict_reg Ebx sat rt tt lasm in
        let (sat, rt, tt, lasm) = evict_reg Edx sat rt tt lasm in

        (* need to take src1 argument before the evictions *)
        let src1_str = get_src_string src1 sat 0 in

        (* get the destination and its string - request ebx *)
        let (sat, rt, tt, lasm, reg) = get_reg_resource dest 
                ltac sat rt tt lasm Ebx in

        (* get the source string for src2 *)
        let src2_str = get_src_string src2 sat 0 in

        (*
         * Virtual 8086 mode displays some goofy divide characteristics
         * when the extended register set is used.  Because of this, to be
         * safe, we will zero out all registers used before beginning.
         *)
        let asm = ("\txor\teax,\teax") in
        let lasm = asm :: lasm in
        let asm = ("\txor\tebx,\tebx") in
        let lasm = asm :: lasm in
        let asm = ("\txor\tedx,\tedx") in
        let lasm = asm :: lasm in
        (*
         * NOTE:  We do not change the sources to reflect where they are moved,
         * which means the registers will remain free.  These types of mul/div 
         * improvements are a stretch goal if time allows.
         *)
        let sz = src1.tds_size in
        if sz == 1 then
                let asm = ("\tmov\tal,\t" ^ src1_str) in
                let lasm = asm :: lasm in
                let asm = ("\tmov\tbl,\t" ^ src2_str) in
                let lasm = asm :: lasm in
                let asm = ("\tdiv\tbl") in
                let lasm = asm :: lasm in
                if op == Tdiv then
                        let asm = ("\tmov\tbl,\tal") in
                        let lasm = asm :: lasm in
                        (lasm, sat, rt, tt, env)
                else (* op == Trdiv *)
                        let asm = ("\tmov\tbl,\tah") in
                        let lasm = asm :: lasm in
                        (lasm, sat, rt, tt, env)
        else if sz == 2 then
                let asm = ("\tmov\tax,\t" ^ src1_str) in
                let lasm = asm :: lasm in
                let asm = ("\tmov\tbx,\t" ^ src2_str) in
                let lasm = asm :: lasm in
                let asm = ("\tdiv\tbx") in
                let lasm = asm :: lasm in
                if op == Tdiv then
                        let asm = ("\tmov\tbx,\tax") in
                        let lasm = asm :: lasm in
                        (lasm, sat, rt, tt, env)
                else (* op == Trdiv *)
                        let asm = ("\tmov\tbx,\tdx") in
                        let lasm = asm :: lasm in
                        (lasm, sat, rt, tt, env)
        else (* sz == 4 *)
                let asm = ("\tmov\teax,\t" ^ src1_str) in
                let lasm = asm :: lasm in
                let asm = ("\tmov\tebx,\t" ^ src2_str) in
                let lasm = asm :: lasm in
                let asm = ("\tdiv\tebx") in
                let lasm = asm :: lasm in
                if op == Tdiv then
                        let asm = ("\tmov\tebx,\teax") in
                        let lasm = asm :: lasm in
                        (lasm, sat, rt, tt, env)
                else (* op == Trdiv *)
                        let asm = ("\tmov\tebx,\tedx") in
                        let lasm = asm :: lasm in
                        (lasm, sat, rt, tt, env)
                

(* 
 * mul_op
 *
 *      Provide assembly for a multiply operation.
 *
 * Supported multiply operations:  mul
 *
 * Intended assembly:
 *      
 *      < evict eax >
 *      < evict edx > ; ensure destination is routed to REGA
 *      mov     REGA,   src1
 *      mul     src2
 *      ; Perform work to get result into eax
 *
 * Output: (lasm, sat, rt, tt, env) *)
let mul_op ltac lasm sat rt tt env = 
        let tac_el = List.hd ltac in
        let op = tac_el.to_op in
        let dest = tac_el.to_dest in
        let src1 = tac_el.to_src1 in
        let src2 = tac_el.to_src2 in
        (* evict the contents of both EAX and EDX *)
        let (sat, rt, tt, lasm) = evict_reg Eax sat rt tt lasm in
        let (sat, rt, tt, lasm) = evict_reg Edx sat rt tt lasm in
        (* get the destination *)
        let (sat, rt, tt, lasm, reg) = get_reg_resource dest 
                ltac sat rt tt lasm Eax in
        (* get the source strings for src1 and src2 *)
        let src1_str = get_src_string src1 sat 0 in
        let src2_str = get_src_string src2 sat 0 in
        (* dest is REGA of some form - generate assembly *)
        let sz = src1.tds_size in
        if sz == 1 then
                let asm = ("\txor\teax,\teax") in
                let lasm = asm :: lasm in
                let asm = ("\tmov\tal,\t" ^ src1_str) in
                let lasm = asm :: lasm in
                let asm = ("\tmov\tdl,\t" ^ src2_str) in
                let lasm = asm :: lasm in
                let asm = ("\tmul\tdl") in
                let lasm = asm :: lasm in
                (lasm, sat, rt, tt, env)
        else if sz == 2 then
                let asm = ("\txor\teax,\teax") in
                let lasm = asm :: lasm in
                let asm = ("\txor\tedx,\tedx") in
                let lasm = asm :: lasm in
                let asm = ("\tmov\tax,\t" ^ src1_str) in
                let lasm = asm :: lasm in
                let asm = ("\tmov\tdx,\t" ^ src2_str) in
                let lasm = asm :: lasm in
                let asm = ("\tmul\tdx") in
                let lasm = asm :: lasm in
                (lasm, sat, rt, tt, env)
        else (* sz == 4 *)
                let asm = ("\txor\teax,\teax") in
                let lasm = asm :: lasm in
                let asm = ("\tmov\teax,\t" ^ src1_str) in
                let lasm = asm :: lasm in
                let asm = ("\tmov\tedx,\t" ^ src2_str) in
                let lasm = asm :: lasm in
                let asm = ("\tmul\tedx") in
                let lasm = asm :: lasm in
                (lasm, sat, rt, tt, env)


(* 
 * assign_op
 *
 *      Provide assembly for an assignment operation.
 *
 * The supported ops are:
 *
 *      mov
 *
 *
 * Intended assembly for binary operations
 *
 *      mov     dest,   src1
 *
 * output: (lasm, sat, rt, tt, env) 
 *)
let assign_op ltac lasm sat rt tt env =
        let tac_el = List.hd ltac in
        let op = tac_el.to_op in
        let dest = tac_el.to_dest in
        let src1 = tac_el.to_src1 in

        (* first get the register for the temporary *) 
        let (sat, rt, tt, lasm, reg) =
               get_reg_resource dest ltac sat rt tt lasm Reg_none in
        let dest_str = string_of_reg reg dest.tds_size in
        (* get the string for the first source operand *)
        let src_str = get_src_string src1 sat 0 in
        (* perform the assignment *)
        let asm = ("\tmov\t" ^ dest_str ^ ",\t" ^ src_str) in
        let lasm = asm :: lasm in
        (lasm, sat, rt, tt, env)


(* 
 * array_op
 *
 *      Provide assembly for an array operation.  This includes
 *      only array accesses, not assignments.
 *
 * The supported ops are:
 *
 *      < array access >
 *
 *
 * Intended assembly for binary operations
 *
 *      mov     edi,    < array index >
 *      add     edi,    < array base offset >
 *      mov     dest,   ds:[edi]
 *
 * output: (lasm, sat, rt, tt, env) 
 *)
let array_op ltac lasm sat rt tt env =
        let tac_el = List.hd ltac in
        let op = tac_el.to_op in
        let dest = tac_el.to_dest in
        let src1 = tac_el.to_src1 in
        let src2 = tac_el.to_src2 in
        (* first get the register for the temporary *) 
        let (sat, rt, tt, lasm, reg) =
               get_reg_resource dest ltac sat rt tt lasm Reg_none in
        let dest_str = string_of_reg reg dest.tds_size in
        (* get the string for the source operands *)
        let src1_str = get_src_string src1 sat 0 in
        (* 
         * source 2 must be considered a 4-byte value if it
         * is in a register - in order to work with edi.
         *)
        let src2_str = get_src_string src2 sat 4 in 
        (* generate the assembly *)
        let asm = ("\tmov\tedi,\t" ^ src2_str) in
        let lasm = asm :: lasm in
        let asm = ("\tadd\tedi,\t" ^ src1_str) in
        let lasm = asm :: lasm in
        let asm = ("\tmov\t" ^ dest_str ^ ",\tds:[edi]") in
        let lasm = asm :: lasm in
        (lasm, sat, rt, tt, env)

(* 
 * array_asn_op
 *
 *      Provide assembly for an array assignment.
 *
 * The supported ops are:
 *
 *      < array assignment >
 *
 *
 * Intended assembly for binary operations
 *
 *      mov     edi,    < array index >
 *      add     edi,    < array base offset >
 *      mov     SIZE ptr ds:[edi],      src
 *
 * output: (lasm, sat, rt, tt, env) 
 *)
let array_asn_op ltac lasm sat rt tt env =
        let tac_el = List.hd ltac in
        let op = tac_el.to_op in
        let arr = tac_el.to_dest in
        let index = tac_el.to_src1 in
        let src = tac_el.to_src2 in
        (* get the string for the source *)
        (* if the source is not in a reg, we need to move it there *)
        (* if the source is an immediate, we need to move it to a reg *)
        let (reg, sat, rt, tt, lasm) = 
                place_in_reg src ltac sat rt tt lasm in
        let src_str = string_of_reg reg src.tds_size in
        (* get the ptr string *)
        let ptr_str = 
                (if arr.tds_value == 1 then ("byte ptr ")
                else if arr.tds_value == 2 then ("word ptr ")
                else if arr.tds_value == 4 then ("dword ptr ")
                else raise (Failure ("invalid array data size: " ^
                        (string_of_int arr.tds_value)))) in
        (* get the string for the array operands *)
        let arr_str = get_src_string arr sat 0 in
        let index_str = get_src_string index sat 4 in
        (* generate the assembly *)
        let asm = ("\tmov\tedi,\t" ^ index_str) in
        let lasm = asm :: lasm in
        let asm = ("\tadd\tedi,\t" ^ arr_str) in
        let lasm = asm :: lasm in
        let asm = ("\tmov\t" ^ ptr_str ^ "ds:[edi],\t" ^ src_str) in
        let lasm = asm :: lasm in
        (lasm, sat, rt, tt, env)

(* 
 * claim_tmp_stack_space
 *
 *      Claim temporary stack space for a TAC operation.  This
 *      will temporarily allocate space on the stack for any 
 *      temporary values that need to be stored within the expression.
 *
 * Output: (sat, tt) 
 *)
let claim_tmp_stack_space sat tt tac_dest =
        if tac_dest.tds_type != Dtmp then (sat, tt)
        else
        let sat_name = sat_name_from_tac_ds tac_dest in
        let sat_symb = 
                (if NameMap.mem sat_name sat then (NameMap.find sat_name sat)
                else 
                        begin
                        let _ = print_endline " *** FAILURE" in
                        let _ = print_sat sat in
                        let _ = print_tac_dest_src tac_dest in
                        raise (Failure ("claim_tmp_stack_space: name not found " ^
                                sat_name))
                        end) in
        (* if the offset is 0, then it is not using space *)
        if sat_symb.sat_soff != 0 then (sat, tt)
        else
        let off = tt.tt_base + (List.hd tt.tt_free) in
        (* update the temporaries table *) 
        let tt = {
                tt_base = tt.tt_base;
                tt_free = (List.tl tt.tt_free) } in
        let new_symb = {
                sat_name = sat_symb.sat_name;
                sat_type = sat_symb.sat_type;
                sat_size = sat_symb.sat_size;
                sat_reg = sat_symb.sat_reg;
                sat_loc = sat_symb.sat_loc;
                sat_soff = off;
                sat_nextuse = sat_symb.sat_nextuse } in
        (* update the sat *)
        let sat = NameMap.remove sat_name sat in
        let sat = NameMap.add sat_name new_symb sat in
        (sat, tt)


(* 
 * release_tmp_stack_space
 *
 *      Release temporary stack space that was previously claimed
 *      for a TAC expression's temporary variables.  This function
 *      checks whether a temporary value is used in the future, and
 *      if not, it releases its stack allocation.
 *
 * Output: (sat, tt, rt) 
 *)
let release_tmp_stack_space sat tt rt ltac src =
        if src.tds_type != Dtmp then (sat, tt, rt)
        else
        let in_future = exists_in_future src.tds_value ltac in
        if in_future == true then (sat, tt, rt)
        else
        (* this source is a tmp that does not exist in future *)
        (* release it, give resources back *)
        let sat_name = sat_name_from_tac_ds src in
        let sat_symb =
               (if NameMap.mem sat_name sat then (NameMap.find sat_name sat)
               else raise (Failure ("release_tmp_stack_space:  Name not found "
               ^ sat_name))) in
        let my_free = sat_symb.sat_soff - tt.tt_base in 
        (* give space back to tt *)
        let tt = {
                tt_base = tt.tt_base;
                tt_free = my_free :: tt.tt_free } in
        (* if I have a reg, release it in rt *)
        let my_reg = sat_symb.sat_reg in
        let rt = (if my_reg == Reg_none then (rt)
                else (update_rt my_reg invld_sat_symb rt true)) in
        (* now update the symbol *)
        let new_symb = {
                sat_name = sat_symb.sat_name;
                sat_type = sat_symb.sat_type;
                sat_size = sat_symb.sat_size;
                sat_reg = Reg_none; (* release the register, regardless *)
                sat_loc = sat_symb.sat_loc;
                sat_soff = 0;
                sat_nextuse = sat_symb.sat_nextuse } in
        let sat = NameMap.remove sat_name sat in
        let sat = NameMap.add sat_name new_symb sat in
        (sat, tt, rt)
         
        

(* 
 * generate_asm_for_op
 *
 *      Generate assembly for a given TAC operation.  This is
 *      the main assembly generating function for handling a list
 *      of TAC operations.
 *
 * Output: (lasm, sat, rt, tt, env) 
 *)
let generate_asm_for_op ltac sat rt tt lasm env =
        let tac_el = List.hd ltac in
        let op = tac_el.to_op in
        let dest = tac_el.to_dest in
        let src1 = tac_el.to_src1 in
        let src2 = tac_el.to_src2 in

        (* we need to get stack space for new tmp variables *)
        let (sat, tt) = claim_tmp_stack_space sat tt dest in

        let (lasm, sat, rt, tt, env) =
        match op with
          Tadd -> (binary_op ltac lasm sat rt tt env)
        | Tsub -> (binary_op ltac lasm sat rt tt env)
        | Tmul -> (mul_op ltac lasm sat rt tt env)
        | Tdiv -> (div_op ltac lasm sat rt tt env)
        | Trdiv -> (div_op ltac lasm sat rt tt env)
        | Txor -> (binary_op ltac lasm sat rt tt env)
        | Tor -> (binary_op ltac lasm sat rt tt env)
        | Tand -> (binary_op ltac lasm sat rt tt env)
        | Teq -> (comparison_op ltac lasm sat rt tt env)
        | Tne -> (comparison_op ltac lasm sat rt tt env)
        | Tlt -> (comparison_op ltac lasm sat rt tt env)
        | Tgt -> (comparison_op ltac lasm sat rt tt env)
        | Tle -> (comparison_op ltac lasm sat rt tt env)
        | Tge -> (comparison_op ltac lasm sat rt tt env)
        | Tuminus -> (unary_op ltac lasm sat rt tt env)
        | Tnot -> (unary_op ltac lasm sat rt tt env)
        | Tjmp -> (jmp_op ltac lasm sat rt tt env)
        | Tjmpe -> (jmp_op ltac lasm sat rt tt env)
        | Tjmpne -> (jmp_op ltac lasm sat rt tt env)
        | Tjmpgt -> (jmp_op ltac lasm sat rt tt env)
        | Tjmplt -> (jmp_op ltac lasm sat rt tt env)
        | Tjmpge -> (jmp_op ltac lasm sat rt tt env)
        | Tjmple -> (jmp_op ltac lasm sat rt tt env)
        | Tlabel -> let asm = ("L" ^ (string_of_int dest.tds_value) ^
                                ":") in
                let lasm = asm :: lasm in
                (lasm, sat, rt, tt, env)
        | Tasn -> (assign_op ltac lasm sat rt tt env)
        | Tarray -> (array_op ltac lasm sat rt tt env)
        | Taasn -> (array_asn_op ltac lasm sat rt tt env)
        | Tcmp -> (* get the string for the first and second source operand *) 
                let src2_str = get_src_string src2 sat 0 in
                (* load a register with the first source *)
                let (reg, sat, rt, tt, lasm) = 
                        place_in_reg src1 ltac sat rt tt lasm in
                let src1_str = get_src_string src1 sat 0 in
                let asm = ("\tcmp\t" ^ src1_str ^ ",\t" ^ src2_str) in
                let lasm = asm :: lasm in
                (lasm, sat, rt, tt, env)  
        | Tret -> (* Note: The source of the return value is actually in dest *)
                let src_str = get_src_string dest sat 0 in
                (* Result will be placed in EAX, regardless of size *)
                let (sat, rt, tt, lasm) = evict_reg Eax sat rt tt lasm in
                (* generate the assembly *)
                let asm = ("\tmov\teax,\t" ^ src_str) in
                let lasm = asm :: lasm in   
                (* the "ret" instruction is handled by the function construct *)
                (lasm, sat, rt, tt, env)
        | Tparam -> (*let (sat, rt, tt, lasm) = evict_reg Eax sat rt tt lasm
                in*)
                (* 
                 * params are only pushed before function calls - we need to
                 * evict all registers before calls, so evict here 
                 *)
                let (sat, rt, tt, lasm) = evict_all_reg sat rt tt lasm in
                (* get the string for the parameter *)
                let p_str = get_src_string dest sat 0 in
                let asm = ("\tmov\teax,\t" ^ p_str) in
                let lasm = asm :: lasm in
                let asm = ("\tpush\teax") in
                let lasm = asm :: lasm in
                (lasm, sat, rt, tt, env) 
        | Tcall -> let (sat, rt, tt, lasm) = evict_all_reg sat rt tt lasm
                in 
                (* value is returned in eax *)
                let (sat, rt, tt, lasm, reg) = get_reg_resource dest 
                        ltac sat rt tt lasm Eax 
                in
                let f_str = src1.tds_name in
                let asm = ("\tcall\t" ^ f_str) in
                let lasm = asm :: lasm in
                (lasm, sat, rt, tt, env)
        | Tevictall -> let (sat, rt, tt, lasm) = evict_all_reg sat rt tt lasm
                in
                (lasm, sat, rt, tt, env)
        | Tespadd -> let imm_str = (string_of_int dest.tds_value) in
                let asm = ("\tadd\tesp,\t" ^ imm_str) in
                let lasm = asm :: lasm in
                (lasm, sat, rt, tt, env)
        | Tevicteax -> let (sat, rt, tt, lasm) = 
                        evict_reg Eax sat rt tt lasm in
                (lasm, sat, rt, tt, env)
        | Tevictebx -> let (sat, rt, tt, lasm) = 
                        evict_reg Ebx sat rt tt lasm in
                (lasm, sat, rt, tt, env)
        | Tevictecx -> let (sat, rt, tt, lasm) = 
                        evict_reg Ecx sat rt tt lasm in
                (lasm, sat, rt, tt, env)
        | Tevictedx -> let (sat, rt, tt, lasm) = 
                        evict_reg Edx sat rt tt lasm in
                (lasm, sat, rt, tt, env)
        in
        (* we need to release stack space for tmp variables *)
        let ltac_tl = List.tl ltac in
        let (sat, tt, rt) = release_tmp_stack_space sat tt rt ltac_tl dest in
        let (sat, tt, rt) = release_tmp_stack_space sat tt rt ltac_tl src1 in
        let (sat, tt, rt) = release_tmp_stack_space sat tt rt ltac_tl src2 in
        (lasm, sat, rt, tt, env)
        



(* 
 * generate_function_assembly
 *
 *      Generate assembly for an entire function.
 *
 * Output: (lasm, env) 
 *)
let generate_function_assembly func lasm gsat env =
        (* setup the portions of the environment which are local *)
        let env = {
                age_implabel = env.age_implabel;
                age_soff = -4; (* locals start at ebp-4 *)
                age_doff = env.age_doff;
                age_nlocals = 0 (* havent added locals yet *) } in
        let fname = func.func_symb.symb_name in
        (* first we need to add the local symbols to the sat *)
        let nmsymb = func.func_locals in
        let lsymb = NameMap.fold (fun k v l -> v :: l) nmsymb [] in
        let (sat, env) = add_local_sat_symbols lsymb gsat env in
        (* grap the list of tac ops *)
        let ltac = func.func_ltac in
        (* initialize rt and tt *)
        let rt = {
                rt_eax = "invld", Rs_free;
                rt_ebx = "invld", Rs_free;
                rt_ecx = "invld", Rs_free;
                rt_edx = "invld", Rs_free;
                rt_nextevict = Eax } in
        (* get the amount of space allocated for temporaries *)
        let tmp_num = get_stack_space_for_temps ltac 0 0 in
        let tmp_space = 4 * tmp_num in (* each tmp gets 4 bytes *)
        let rec get_tt_free numtmps nextoffset freelist =
                if numtmps == 0 then List.rev freelist
                else get_tt_free (numtmps - 1) (nextoffset - 4) 
                        (nextoffset :: freelist)
        in
        let ttfree = get_tt_free tmp_num 0 [] in
        let local_space = 4 * env.age_nlocals in
        let ttbase = -4 - local_space in
        let tt = {
                tt_base = ttbase;
                tt_free = ttfree } in
        (* add the tmp variables to the SAT *)
        let sat = add_tmp_sat_symbols sat tt ltac in

        (* now we add the function name *)
        let asm = (fname ^ " PROC NEAR") in
        let lasm = asm :: lasm in
        let asm = ("\tpush\tebp") in
        let lasm = asm :: lasm in
        let asm = ("\tmov\tebp,\tesp") in
        let lasm = asm :: lasm in
        let asm = ("\tsub\tesp,\t" ^ (string_of_int (local_space + tmp_space)))
        in
        let lasm = asm :: lasm in
        (* now we initialize the locals *)
        let lasm = generate_local_init_assembly sat lsymb lasm in
        (* generate assembly for the Tac elements within the function *)
        let rec gen_all_asm ltac sat rt tt lasm env =
                if ltac == [] then (lasm, sat, rt, tt, env)
                else
                        begin
                        let (lasm, sat, rt, tt, env) = generate_asm_for_op 
                                ltac sat rt tt lasm env in
                        (gen_all_asm (List.tl ltac) sat rt tt lasm env)
                        end
        in
        let (lasm, sat, rt, tt, env) = gen_all_asm ltac sat rt tt lasm env in
        (* we should now be complete with generating assembly for the ops *)
        let asm = ("\tadd\tesp,\t" ^ (string_of_int (local_space + tmp_space)))
        in
        let lasm = asm :: lasm in
        let asm = ("\tpop\tebp") in 
        let lasm = asm :: lasm in
        let asm = ("\tret") in
        let lasm = asm :: lasm in
        (* Final addition - the end of the proc *)
        let asm = (fname ^ " ENDP") in
        let lasm = asm :: lasm in
        (lasm, env)


(* 
 * generate_rcb_assembly
 *
 *      Generate assembly for an entire RCB.
 *
 * Output: (lasm, env) 
 *)
let generate_rcb_assembly rcb lasm gsat env =
        let rname = rcb.rcb_symb.symb_name in
        (* first we need to add the local symbols to the sat *)
        let nmsymb = rcb.rcb_locals in
        let lsymb = NameMap.fold (fun k v l -> v :: l) nmsymb [] in
        let (sat, env) = add_local_sat_symbols lsymb gsat env in
        (* grap the list of tac ops *)
        let ltac = rcb.rcb_ltac in
        (* initialize rt and tt *)
        let rt = {
                rt_eax = "invld", Rs_free;
                rt_ebx = "invld", Rs_free;
                rt_ecx = "invld", Rs_free;
                rt_edx = "invld", Rs_free;
                rt_nextevict = Eax } in
        (* get the amount of space allocated for temporaries *)
        let tmpmap = IntMap.empty in
        let tmp_num = get_stack_space_for_temps ltac 0 0 in
        let tmp_space = 4 * tmp_num in (* each tmp gets 4 bytes *)
        let rec get_tt_free numtmps nextoffset freelist =
                if numtmps == 0 then List.rev freelist
                else get_tt_free (numtmps - 1) (nextoffset - 4) 
                        (nextoffset :: freelist)
        in
        let ttfree = get_tt_free tmp_num 0 [] in
        let local_space = 4 * env.age_nlocals in
        let ttbase = -4 - local_space in
        let tt = {
                tt_base = ttbase;
                tt_free = ttfree } in
        (* add the tmp variables to the SAT *)
        let sat = add_tmp_sat_symbols sat tt ltac in
        (* now we add the RCB name *)
        let asm = (rname ^ " PROC NEAR") in
        let lasm = asm :: lasm in
        let asm = ("\tpush\tebp") in
        let lasm = asm :: lasm in
        let asm = ("\tmov\tebp,\tesp") in
        let lasm = asm :: lasm in
        let asm = ("\tsub\tesp,\t" ^ (string_of_int (local_space + tmp_space)))
        in
        let lasm = asm :: lasm in
        (* now we initialize the locals *)
        let lasm = generate_local_init_assembly sat lsymb lasm in
        (* generate assembly for the Tac elements within the RCB *)
        let rec gen_all_asm ltac sat rt tt lasm env =
                if ltac == [] then (lasm, sat, rt, tt, env)
                else
                        begin
                        let (lasm, sat, rt, tt, env) = generate_asm_for_op 
                                ltac sat rt tt lasm env in
                        (gen_all_asm (List.tl ltac) sat rt tt lasm env)
                        end
        in
        let (lasm, sat, rt, tt, env) = gen_all_asm ltac sat rt tt lasm env in
        (* we should now be complete with generating assembly for the ops *)
        let asm = ("\tadd\tesp,\t" ^ (string_of_int (local_space + tmp_space)))
        in
        let lasm = asm :: lasm in
        let asm = ("\tpop\tebp") in 
        let lasm = asm :: lasm in
        let asm = ("\tret") in
        let lasm = asm :: lasm in
        (* Final addition - the end of the proc *)
        let asm = (rname ^ " ENDP") in
        let lasm = asm :: lasm in
        (lasm, env)


(* 
 * generate_assembly_program
 *
 *      Generate assembly for an entire program.
 *
 * Output:  (lasm) 
 *)
let generate_assembly_program prog =
       (* first lets add the structure surrounding the assembly program *)
       (* 
        * Intended assembly:
        * 
        * ; Begin assembly
        * .386
        *
        * DSEG  SEGMENT PUBLIC
        *       BYTE    3FF0h   dup     (0)
        *       TABLE   DB      '0123456789ABCDEF'
        * DSEG  ENDS
        *
        * SSEG  SEGMENT STACK
        *       BYTE    1000h   dup     (0)
        * SSEG  ENDS
        *
        * CSEG  SEGMENT PUBLIC  USE16
        *       ASSUME  CS:CSEG, DS:DSEG, SS:SSEG
        *
        * START:
        *       mov     ax,     DSEG
        *       mov     ds,     ax
        *
        *       call    main    ; main must be defined
        *
        * EXIT: mov     ah,     4Ch
        *       int     21h
        *
        * < assembly from Functions and RCBs >
        *
        * < assembly for printing numbers >
        *
        * CSEG  ENDS
        *
        * END   START
        * ; End assembly
        *)
       let lasm = [] in
       let asm = ("; Begin Assembly") in
       let lasm = asm :: lasm in
       let asm = (".386") in
       let lasm = asm :: lasm in
       let asm = ("\n") in
       let lasm = asm :: lasm in
       let asm = ("DSEG\tSEGMENT\tPUBLIC") in
       let lasm = asm :: lasm in
       let asm = ("\tBYTE\t3FF0h\tdup\t(0)") in
       let lasm = asm :: lasm in
       let asm = ("\tTABLE\tDB\t'0123456789ABCDEF'") in
       let lasm = asm :: lasm in
       let asm = ("DSEG\tENDS") in
       let lasm = asm :: lasm in
       let asm = ("\n") in
       let lasm = asm :: lasm in
       let asm = ("SSEG\tSEGMENT\tSTACK") in
       let lasm = asm :: lasm in
       let asm = ("\tBYTE\t1000h\tdup\t(0)") in
       let lasm = asm :: lasm in
       let asm = ("SSEG\tENDS") in
       let lasm = asm :: lasm in
       let asm = ("\n") in
       let lasm = asm :: lasm in
       let asm = ("CSEG\tSEGMENT\tPUBLIC\tUSE16") in
       let lasm = asm :: lasm in
       let asm = ("\tASSUME\tCS:CSEG, DS:DSEG, SS:SSEG") in
       let lasm = asm :: lasm in
       let asm = ("\n") in
       let lasm = asm :: lasm in

       (* ENTRY POINT *)
       let asm = "START:" in
       let lasm = asm :: lasm in
       let asm = "\tmov\tax,\tDSEG" in
       let lasm = asm :: lasm in
       let asm = "\tmov\tds,\tax" in
       let lasm = asm :: lasm in
       let asm = "\tcall\tmain" in
       let lasm = asm :: lasm in
       let asm = "EXIT:\tmov\tah,\t4Ch" in
       let lasm = asm :: lasm in
       let asm = "\tint\t21h" in
       let lasm = asm :: lasm in
       let asm = "\n" in
       let lasm = asm :: lasm in


       (* setup the initial environment *)
       let env = {
               age_implabel = 0;
               age_soff = 0;
               age_doff = 0;
               age_nlocals = 0 } in

       (* create the global sat *)
       let gsat = Tac_types.NameMap.empty in
       let lsymb = NameMap.fold (fun k v l -> v :: l) prog.prog_globals [] in
       let (gsat, env) = add_global_sat_symbols lsymb gsat env in

       (* generate the assembly for functions *)
       (* Output: (lasm, env) *)
       let rec gen_all_funcs lfunc lasm gsat env = 
               if lfunc == [] then (lasm, env)
               else
                       begin
                       let func = List.hd lfunc in
                       let lfunc = List.tl lfunc in
                       let (lasm, env) = generate_function_assembly 
                                func lasm gsat env in
                       let asm = ("\n") in
                       let lasm = asm :: lasm in
                       (gen_all_funcs lfunc lasm gsat env)
                       end
       in
       let (lasm, env) = gen_all_funcs prog.prog_lfunc lasm gsat env in
       (* generate the assembly for RCBs *)
       let rec gen_all_rcbs lrcb lasm gsat env = 
               if lrcb == [] then (lasm, env)
               else
                       begin
                       let rcb = List.hd lrcb in
                       let lrcb = List.tl lrcb in
                       let (lasm, env) = generate_rcb_assembly 
                                rcb lasm gsat env in
                       let asm = ("\n") in
                       let lasm = asm :: lasm in
                       (gen_all_rcbs lrcb lasm gsat env)
                       end
       in
       let (lasm, env) = gen_all_rcbs prog.prog_lrcb lasm gsat env in

       (* add the printing functionality *)
       let lasm = hex_print lasm in

       let asm = ("CSEG\tENDS") in
       let lasm = asm :: lasm in
       let asm = ("\n") in
       let lasm = asm :: lasm in
       let asm = ("END\tSTART") in
       let lasm = asm :: lasm in
       let asm = ("; End Assembly") in
       let lasm = asm :: lasm in
       (lasm)
