(* 
 * tac_display.ml 
 *)

(*
 * Routines to display three address code in a user friendly form.
 * This is intended only for debug.
 *)

open Tac_types
open Ast


let debug_print = true

(*
 * Return a string representing the TAC Op
 *)

(* 
 * Return a string from a Three Address Code Operation 
 *)
let string_from_tac op =
        match op with
            Tadd -> "Add"
          | Tsub -> "Sub"
          | Tmul -> "Mul"
          | Tdiv -> "Div"
          | Trdiv -> "RDiv"
          | Txor -> "Xor"
          | Tor -> "Or"
          | Tand -> "And"
          | Teq -> "Equal"
          | Tne -> "NEqual"
          | Tlt -> "Less"
          | Tgt -> "Greater"
          | Tle -> "LTE"
          | Tge -> "GTE"
          | Tuminus -> "Uminus"
          | Tnot -> "Not"
          | Tjmp -> "Jmp"
          | Tjmpe -> "Jmpe"
          | Tjmpne -> "Jmpne"
          | Tjmpgt -> "Jmpgt"
          | Tjmplt -> "Jmplt"
          | Tjmpge -> "Jmpge"
          | Tjmple -> "Jmple"
          | Tlabel -> "Label"
          | Tarray -> "Array"
          | Tasn -> "Assign"
          | Taasn -> "ArrAsn"
          | Tcmp -> "Cmp"
          | Tret -> "Ret"
          | Tparam -> "Param"
          | Tcall -> "Call"
          | Tespadd -> "esp add"
          | Tevictall -> "evict all"
          | Tevicteax -> "evict eax"
          | Tevictebx -> "evict ebx"
          | Tevictecx -> "evict ecx"
          | Tevictedx -> "evict edx"

(*
 * Return a string format of a dest / src
 *)
let string_from_dest_src ds =
        if ds.tds_type == Dvar then
                ("Var = " ^ ds.tds_name)
        else if ds.tds_type == Darray then
                ("Array = " ^ ds.tds_name ^ "[" ^ (string_of_int ds.tds_size) ^ "]") 
        else if ds.tds_type == Dimm then
                ("Imm = " ^ (string_of_int ds.tds_value))
        else
                ("TRslt = " ^ (string_of_int ds.tds_value))

(*
 * Get string from a data_type
 *)
let string_from_data_type dt =
        match dt with
            Void -> "Void"
          | Char -> "Char"
          | Ord1 -> "Ord1"
          | Ord2 -> "Ord2"
          | Ord4 -> "Ord4"

(*
 * Return a string from a TAC dest/src type
 *)
let string_from_tds_type tdstype =
        match tdstype with
           Dvar -> "Dvar"
         | Darray -> "Darray"
         | Dimm -> "Dimm"
         | Dtmp -> "Dtmp"

(*
 * Print a TAC Element
 *)
let print_tac_dest_src dest_src =
        print_endline ("\t\tTYPE: " ^ 
        (string_from_tds_type dest_src.tds_type) ^ 
        "    NAME: " ^ dest_src.tds_name ^
        "    VALUE: " ^ (string_of_int dest_src.tds_value) ^
        "    SIZE: " ^ (string_of_int dest_src.tds_size))

(*
 * Print a single Three Address Code Element
 *)
let print_tac_el tac =
        print_endline ("\t< " ^ (string_from_tac tac.to_op) ^ ", " ^
        (string_from_dest_src tac.to_dest) ^ ", " ^
        (string_from_dest_src tac.to_src1) ^ ", " ^
        (string_from_dest_src tac.to_src2) ^ " >")

(*
 * Print a single group Three Address Code Element
 *)
let print_gtac_el gtac =
        let g = fst gtac in
        let tac = snd gtac in
        print_endline ("\t( " ^ (string_of_int g) ^
        ", < " ^ (string_from_tac tac.to_op) ^ ", " ^
        (string_from_dest_src tac.to_dest) ^ ", " ^
        (string_from_dest_src tac.to_src1) ^ ", " ^
        (string_from_dest_src tac.to_src2) ^ " > )")

(*
 * Print information on a Three Address Code element in the 
 * case of an error.
 *)
let print_tac_error tac = 
        print_endline "\tPRINT_TAC_ERROR";
        print_tac_el tac;
        print_endline "\tDESTINATION:";
        print_tac_dest_src tac.to_dest;
        print_endline "\tSOURCE 1:";
        print_tac_dest_src tac.to_src1;
        print_endline "\tSOURCE 2:";
        print_tac_dest_src tac.to_src2

(*
 * Print a list of Three Address Code elements
 *)
let rec print_tac_list tl =
        if tl == [] then print_endline "ENDFUNC"
        else
                begin 
                let el = (List.hd tl) in
                print_tac_el el;
                print_tac_list (List.tl tl)
                end

(*
 * Print a list of grouped Three Address Code elements
 *)
let rec print_gtac_list gtl =
        if gtl == [] then print_endline "ENDFUNC"
        else
                begin 
                let el = (List.hd gtl) in
                print_gtac_el el;
                print_gtac_list (List.tl gtl)
                end

(*
 * Print Three Address Code information for a function
 *)
let print_tac_func f =
        let fs = f.func_symb in
        print_endline ("FUNC:  " ^ (string_from_data_type fs.symb_dtype) ^ 
        " " ^ fs.symb_name);
        print_tac_list f.func_ltac

(*
 * Print Three Address Code information for a Random Code Block
 *)
let print_tac_rcb r =
        let rs = r.rcb_symb in
        print_endline ("RCB:  " ^ rs.symb_name);
        print_tac_list r.rcb_ltac

(*
 * Print Three Address Code information for an entire program
 *)
let print_tac_program p =
        List.map (fun x-> print_tac_func x) p.prog_lfunc;
        List.map (fun x -> print_tac_rcb x) p.prog_lrcb
        
(*
 * Debug Prints
 *)

(*
 * Print grouped TAC information if debug_print is true
 *)
let debug_print_gtacl gtac_list =
        if debug_print then
                print_gtac_list (List.rev gtac_list)
