(*
 * tacgen.ml
 *)

open Ast
open Tac_types
open Tac_display

(* 
 * Routines to generate Three Address Code
 *)
        
(****************************************************************
 *
 * Routines for environment maintainence and sizing
 *
 ****************************************************************)

(*
 * get_seed
 *
 *      Get a seed value for an RCB call.  This both updates the
 *      "random" seed for the current call and provides it to the
 *      calling function.  Seed values are tracked in the environment.
 *
 * Output:  (env, seed)
 *)
let get_seed env =
        let next_seed = (((env.te_seed * 4297) + 391) mod 100) in
        let env = {
                te_label = env.te_label;
                te_trslt = env.te_trslt;
                te_seed = next_seed;
                te_group = env.te_group } in
        (env, next_seed)

(*
 * get_size_from_data_type
 *
 *      Get an integer representation of a data type size from an 
 *      input MOC-V supported data type.
 *
 * Output:  (size)
 *)
let get_size_from_data_type dt =
       match dt with
           Char -> 1
         | Ord1 -> 1
         | Ord2 -> 2
         | Ord4 -> 4
         | Void -> 0

(****************************************************************
 *
 * Routines to Generate Symbol Tables
 *
 ****************************************************************)

(*
 * populate_vdecl_symbols
 *
 *      Populate a symbols map given a list of variable declarations
 *      provided by the parser.  The symbols map is a mapping between
 *      symbol name and symbols information, provided in the form
 *      of the "symbol" type (see tac_types.ml).
 *
 * Output:  (symbol map)
 *)
let rec populate_vdecl_symbols list_in smap =
        if list_in == [] then smap
        else let curr_vdecl = List.hd list_in in
        match curr_vdecl with 
            Initvardecl(dt, n, i) ->
                if NameMap.mem n smap then
                       raise (Failure ("variable already exists " ^ n))
                else
                  let curr_size = get_size_from_data_type dt in 
                  let curr_symb = { symb_name = n;
                                    symb_stype = Svar;
                                    symb_dtype = dt;
                                    symb_offset = 0;
                                    symb_init = i;
                                    symb_size = curr_size;
                                    symb_isparam = false;
                                    symb_paramnum = 0; 
                                    symb_paramsizes = [] } in
                  let smap = NameMap.add n curr_symb smap in
                  populate_vdecl_symbols (List.tl list_in) smap
          | Uninitvardecl(dt, n) ->
                if NameMap.mem n smap then
                        raise (Failure ("variable already exists " ^ n))
                else
                  let curr_size = get_size_from_data_type dt in
                  let curr_symb = { symb_name = n;
                                    symb_stype = Svar;
                                    symb_dtype = dt;
                                    symb_offset = 0;
                                    symb_init = 0;
                                    symb_size = curr_size;
                                    symb_isparam = false;
                                    symb_paramnum = 0; 
                                    symb_paramsizes = [] } in
                  let smap = NameMap.add n curr_symb smap in
                  populate_vdecl_symbols (List.tl list_in) smap
          | Initarraydecl(n, s, i) ->
                if NameMap.mem n smap then
                        raise (Failure ("variable already exists " ^ n))
                else
                  let curr_symb = { symb_name = n;
                                    symb_stype = Sarray;
                                    symb_dtype = Ord1;
                                    symb_offset = 0;
                                    symb_init = i;
                                    symb_size = s;
                                    symb_isparam = false;
                                    symb_paramnum = 0;
                                    symb_paramsizes = [] } in
                  let smap = NameMap.add n curr_symb smap in
                  populate_vdecl_symbols (List.tl list_in) smap
          | Uninitarraydecl(n, s) ->
                if NameMap.mem n smap then
                        raise (Failure ("variable already exists " ^ n))
                else
                  let curr_symb = { symb_name = n;
                                    symb_stype = Sarray;
                                    symb_dtype = Ord1;
                                    symb_offset = 0;
                                    symb_init = 0;
                                    symb_size = s;
                                    symb_isparam = false;
                                    symb_paramnum = 0; 
                                    symb_paramsizes = [] } in
                  let smap = NameMap.add n curr_symb smap in
                  populate_vdecl_symbols (List.tl list_in) smap

(*
 * populate_func_symbols
 *
 *      Populate a symbols map with function symbols.  These symbols
 *      follow the same format as other variables, and are provided 
 *      by the list of function declarations from the parser.
 *
 * Output:  (symbol map)
 *)
let rec populate_func_symbols list_in smap =
        if list_in == [] then smap
        else let curr_fdecl = List.hd list_in in
        if NameMap.mem curr_fdecl.fdecl_name smap then
                raise(Failure ("function name already exists " ^ curr_fdecl.fdecl_name))
        else 
                begin
                let get_size_from_param p =
                        match p with
                            Pdecl(dt, n) -> (get_size_from_data_type dt)
                          | Pdecla(n, s) -> s
                in 
                let rec get_param_list_sizes pl sl =
                        if pl == [] then sl
                        else 
                                begin
                                let cp = List.hd pl in
                                let pl = List.tl pl in
                                let sl = (get_size_from_param cp) :: sl in
                                get_param_list_sizes pl sl
                                end
                in
                let pss = get_param_list_sizes curr_fdecl.fdecl_params [] in
                let curr_symb = { symb_name = curr_fdecl.fdecl_name;
                               symb_stype = Sfunc;
                               symb_dtype = curr_fdecl.fdecl_rtype;
                               symb_offset = 0;
                               symb_init = 0;
                               symb_size = 0;
                               symb_isparam = false;
                               symb_paramnum = 0;
                               symb_paramsizes = (List.rev pss) } in
                let smap = NameMap.add curr_fdecl.fdecl_name curr_symb smap in
                populate_func_symbols (List.tl list_in) smap
                end

(*
 * populate_rcb_symbols
 *
 *      Populate a symbols map with RCB symbol information provided
 *      by the parser.
 *
 * Output:  (symbol map)
 *)
let rec populate_rcb_symbols list_in smap =
        if list_in == [] then smap
        else let curr_rcb = List.hd list_in in
        if NameMap.mem curr_rcb smap then
                raise(Failure ("rcb name already exists " ^ curr_rcb))
        else let curr_symb = { symb_name = curr_rcb;
                               symb_stype = Srcb;
                               symb_dtype = Void;
                               symb_offset = 0;
                               symb_init = 0;
                               symb_size = 0;
                               symb_isparam = false;
                               symb_paramnum = 0;
                               symb_paramsizes = [] } in
        let smap = NameMap.add curr_rcb curr_symb smap in
        populate_rcb_symbols (List.tl list_in) smap

(*
 * populate_param_symbols
 *
 *      Populate a symbols map with symbol information from function
 *      parameters provided by the parser.
 *
 * Output:  (symbol map)
 *)
let rec populate_param_symbols list_in smap paramnum =
        if list_in == [] then smap
        else let curr_param = List.hd list_in in
        match curr_param with
            Pdecl(dt, n) ->
                    if NameMap.mem n smap then
                            raise (Failure ("param name already exists " ^ n))
                    else let curr_size = get_size_from_data_type dt in
                    let curr_symb = { symb_name = n;
                                      symb_stype = Svar;
                                      symb_dtype = dt;
                                      symb_offset = 0;
                                      symb_init = 0;
                                      symb_size = curr_size;
                                      symb_isparam = true;
                                      symb_paramnum = paramnum;
                                      symb_paramsizes = [] } in
                    let smap = NameMap.add n curr_symb smap in
                    populate_param_symbols (List.tl list_in) smap (paramnum - 1)
          | Pdecla(n, s) -> 
                    if NameMap.mem n smap then
                            raise (Failure ("param array name already exists " ^ n))
                    else let curr_symb = { symb_name = n;
                                           symb_stype = Sarray;
                                           symb_dtype = Ord1;
                                           symb_offset = 0;
                                           symb_init = 0;
                                           symb_size = s;
                                           symb_isparam = true;
                                           (* these don't go on the stack *)
                                           symb_paramnum = 0; 
                                           symb_paramsizes = [] } in
                    let smap = NameMap.add n curr_symb smap in
                    populate_param_symbols (List.tl list_in) smap paramnum

                                           

(*
 * populate_symbols
 *
 *      Populate all symbols from a program given a list of variable
 *      declarations, a list of function declarations, a list of 
 *      RCB declarations, and a list of parameters.  This routine can
 *      be called for populating both global and local symbol maps.
 *
 * Output:  (symbol map)
 *)             
let rec populate_symbols lvdecl_in lfdecl_in lrcbdecl_in params_in smap =
        if lvdecl_in != [] then
                let smap = populate_vdecl_symbols lvdecl_in smap in
                populate_symbols [] lfdecl_in lrcbdecl_in params_in smap
        else if lfdecl_in != [] then
                let smap = populate_func_symbols lfdecl_in smap in
                populate_symbols lvdecl_in [] lrcbdecl_in params_in smap
        else if lrcbdecl_in != [] then
                let smap = populate_rcb_symbols lrcbdecl_in smap in
                populate_symbols lvdecl_in lfdecl_in [] params_in smap
        else if params_in != [] then
                let paramnum = List.length params_in in
                let smap = populate_param_symbols params_in smap paramnum in
                populate_symbols lvdecl_in lfdecl_in lrcbdecl_in [] smap
        else smap
        

(****************************************************************
 *
 * Routines for Modifying Three Address Code Structures
 *
 ****************************************************************)

(*
 * append_tac_list
 *
 *      Append a Three Address Code element to a list of such
 *      elements.
 *
 * Output:  (ltac)
 *)
let append_tac_list op d s1 s2 tac_list = 
        let tac_el = {
                to_op = op;
                to_dest = d;
                to_src1 = s1;
                to_src2 = s2 } in
        tac_el :: tac_list

(*
 * append_tac_list_g
 *
 *      Append a Three Address Code element with grouping information
 *      to a list of such elements.
 *
 * Output:  (lgtac)
 *)
let append_tac_list_g g op d s1 s2 tac_list_g =
        let tac_el = {
                to_op = op;
                to_dest = d;
                to_src1 = s1;
                to_src2 = s2 } in
        (g, tac_el) :: tac_list_g

(*
 * get_label_num
 *
 *      Get a unique label number.
 *
 * Output:  (label number, env)
 *)
let get_label_num env =
        let lab = env.te_label in
        let next_env = {
                te_label = env.te_label + 1;
                te_trslt = env.te_trslt;
                te_seed = env.te_seed;
                te_group = env.te_group } in
        (lab, next_env)

(*
 * get_label_dest
 *
 *      Get a TAC dest/src type for a label based on the 
 *      current environment.  This supplies the caller with 
 *      a TAC dest/src specifying a unique label value.
 *
 * Output:  (label dest/src, env)
 *)
let get_label_dest env =
        let (num, env) = get_label_num env in
        let label_dest = {
                tds_type = Dimm;
                tds_name = "invld";
                tds_value = num;
                tds_size = 0 } in
        (label_dest, env)

(*
 * get_trslt_num
 *
 *      Get a unique number for a temporary result.
 *
 * Output:  (trslt number)
 *)
let get_trslt_num env =
        let tr = env.te_trslt in
        let next_env = {
                te_label = env.te_label;
                te_trslt = env.te_trslt + 1; 
                te_seed = env.te_seed;
                te_group = env.te_group } in
        (tr, next_env)

(*
 * get_trslt_dest
 *
 *      Get a destination representing a unique temporary result.
 *
 * Output:  (trslt dest/src)
 *)
let get_trslt_dest env sz =
        let (num, env) = get_trslt_num env in
        let trslt_dest = {
                tds_type = Dtmp;
                tds_name = "invld";
                tds_value = num;
                tds_size = sz } in
        (trslt_dest, env)


(****************************************************************
 *
 * Routines for Type Checking - Which in MOC-V is size checking
 *
 ****************************************************************)

(*
 * tac_get_expr_size_for_group
 *
 *      Determine an expression size for a particular group.  Every
 *      group identifies an expression, and every element within a
 *      single expression should have the same size.  This function
 *      determines what that size should be based on the variables
 *      contained within the expression.
 *
 * Output:  (size)
 *)
let rec tac_get_expr_size_for_group group tac_list =
        if tac_list == [] then 4 (* no sized variables, leave it 4 *)
        else if fst (List.hd tac_list) != group then 
                (tac_get_expr_size_for_group group (List.tl tac_list))
        else
                begin
                let tace = snd (List.hd tac_list) in
                let op = tace.to_op in
                let d = tace.to_dest in
                let s1 = tace.to_src1 in
                let s2 = tace.to_src2 in
                if (op == Tadd || op == Tsub || op == Tmul || op == Tdiv ||
                   op == Trdiv || op == Txor || op == Tor || op == Tand ||
                   op == Teq || op == Tne || op == Tlt || op == Tgt ||
                   op == Tle || op == Tge) then
                        (* 
                         * These are binary ops with the format:
                         *
                         *   < op, dest, src1, src2 >
                         *)
                        if d.tds_type == Dvar then
                                (* variable - sizing must hold *)
                                (d.tds_size)
                        else if d.tds_type == Darray then
                                (* array - sizing must also hold *)
                                (d.tds_value)  (* array access size *)
                        else if s1.tds_type == Dvar then
                                (s1.tds_size)
                        else if s1.tds_type == Darray then
                                (s1.tds_value)
                        else if s2.tds_type == Dvar then
                                (s2.tds_size)
                        else if s2.tds_type == Darray then
                                (s2.tds_value)
                        else (tac_get_expr_size_for_group group (List.tl tac_list))
                else if (op == Tuminus || op == Tnot || op == Tasn) then
                        (*
                         * These are unary ops with the format:
                         *
                         *      < op, dest, src, invld >
                         *)
                        if d.tds_type == Dvar then
                                (d.tds_size)
                        else if d.tds_type == Darray then
                                (d.tds_value)
                        else if s1.tds_type == Dvar then
                                (s1.tds_size)
                        else if s1.tds_type == Darray then
                                (s1.tds_value)
                        else (tac_get_expr_size_for_group group (List.tl tac_list))
                else if (op == Taasn) then
                        (* 
                         * This is an array assignment with format:
                         *
                         *      < op, dest, invld, src >
                         *)
                        if d.tds_type != Darray then raise (Failure ("Taasn op " ^
                                "lacks array dest!"))
                        else (d.tds_value)
                else    (* Not an interesting tac op for sizing, continue *)
                        (tac_get_expr_size_for_group group (List.tl tac_list))
                end

(*
 * make_dest_src_size
 *
 *      Take an input TAC dest/src type and make it the specified
 *      size.
 *
 * Output:  (TAC dest/src)
 *)
let make_dest_src_size dest_src size =
        let dstype = dest_src.tds_type in
        let dsname = dest_src.tds_name in
        let dssize = 
                (if dstype == Darray then dest_src.tds_size
                else size) in
        let dsvalue = 
                (if dstype == Darray then size
                else dest_src.tds_value) in
        let out = {
                tds_type = dstype;
                tds_name = dsname;
                tds_size = dssize;
                tds_value = dsvalue } in
        (out)

(*
 * tac_populate_group_sizes
 *
 *      Take a list of grouped TAC elements and perform the size
 *      checking for each group within.  Also, size elements with
 *      flexible sizes within the group, such as immediates and
 *      temporaries.
 *
 * Output:  (lgtac)
 *)
let rec tac_populate_group_sizes group size lgtac_in lgtac_out =
        if lgtac_in == [] then (lgtac_out)
        else
        let gtac_in = List.hd lgtac_in in
        if (fst gtac_in) != group then
                begin
                (* this is not the same group *)
                let lgtac_out = gtac_in :: lgtac_out in
                let lgtac_in = List.tl lgtac_in in
                (tac_populate_group_sizes group size lgtac_in lgtac_out)
                end
        else
                begin
                (* this is the same group - check the op *)
                let tac = snd gtac_in in
                let op = tac.to_op in
                let d = tac.to_dest in
                let s1 = tac.to_src1 in
                let s2 = tac.to_src2 in
                if (op == Tadd || op == Tsub || op == Tmul || op == Tdiv ||
                   op == Trdiv || op == Txor || op == Tor || op == Tand ||
                   op == Teq || op == Tne || op == Tlt || op == Tgt ||
                   op == Tle || op == Tge) then
                        begin
                        (* < op, dest, src, src > *)
                        let dsize = 
                                (if (d.tds_type == Darray) then
                                        d.tds_value else d.tds_size) in
                        let s1size = 
                                (if (s1.tds_type == Darray) then
                                        s1.tds_value else s1.tds_size) in
                        let s2size = 
                                (if (s2.tds_type == Darray) then
                                        s2.tds_value else s2.tds_size) in
                        if ((d.tds_type == Darray || d.tds_type == Dvar) &&
                           (dsize != size)) then
                                let _ = Tac_display.print_tac_error tac in
                                raise (Failure ("size mismatch: expected size " ^
                                        (string_of_int size)))
                        else if ((s1.tds_type == Darray || s1.tds_type == Dvar) &&
                           (s1size != size)) then
                                let _ = Tac_display.print_tac_error tac in
                                raise (Failure ("size mismatch: expected size " ^
                                        (string_of_int size)))
                        else if ((s2.tds_type == Darray || s2.tds_type == Dvar) &&
                           (s2size != size)) then
                                let _ = Tac_display.print_tac_error tac in
                                raise (Failure ("size mismatch: expected size " ^
                                        (string_of_int size)))
                        else
                                (* all sizes checked out *)
                                let d = make_dest_src_size d size in
                                let s1 = make_dest_src_size s1 size in
                                let s2 = make_dest_src_size s2 size in
                                let tac = {
                                        to_op = op;
                                        to_dest = d;
                                        to_src1 = s1;
                                        to_src2 = s2 } in
                                let gtac = (group, tac) in
                                let lgtac_out = gtac :: lgtac_out in
                                let lgtac_in = List.tl lgtac_in in
                                (tac_populate_group_sizes group size lgtac_in
                                        lgtac_out)
                        end
                else if (op == Tuminus || op == Tnot || op == Tasn) then
                        begin
                        (* < op, dest, src, inv > *)
                        let dsize = 
                                (if (d.tds_type == Darray) then
                                        d.tds_value else d.tds_size) in
                        let s1size = 
                                (if (s1.tds_type == Darray) then
                                        s1.tds_value else s1.tds_size) in
                        if ((d.tds_type == Darray || d.tds_type == Dvar) &&
                           (dsize != size)) then
                                let _ = Tac_display.print_tac_error tac in
                                raise (Failure ("size mismatch: expected size " ^
                                        (string_of_int size)))
                        else if ((s1.tds_type == Darray || s1.tds_type == Dvar) &&
                           (s1size != size)) then
                                let _ = Tac_display.print_tac_error tac in
                                raise (Failure ("size mismatch: expected size " ^
                                        (string_of_int size)))
                        else
                                (* all sizes checked out *)
                                let d = make_dest_src_size d size in
                                let s1 = make_dest_src_size s1 size in
                                let tac = {
                                        to_op = op;
                                        to_dest = d;
                                        to_src1 = s1;
                                        to_src2 = s2 } in
                                let gtac = (group, tac) in
                                let lgtac_out = gtac :: lgtac_out in
                                let lgtac_in = List.tl lgtac_in in
                                (tac_populate_group_sizes group size lgtac_in
                                        lgtac_out)
                        end
                else if (op == Taasn) then
                        begin
                        (* < op, array, inv, src > - inv == don't care *)
                        let dsize = 
                                (if (d.tds_type == Darray) then
                                        d.tds_value else d.tds_size) in
                        let s2size = 
                                (if (s2.tds_type == Darray) then
                                        s2.tds_value else s2.tds_size) in
                        if ((d.tds_type == Darray || d.tds_type == Dvar) &&
                           (dsize != size)) then
                                let _ = Tac_display.print_tac_error tac in
                                raise (Failure ("size mismatch"))
                        else if ((s2.tds_type == Darray || s2.tds_type == Dvar) &&
                           (s2size != size)) then
                                let _ = Tac_display.print_tac_error tac in
                                raise (Failure ("size mismatch"))
                        else
                                (* all sizes checked out *)
                                let d = make_dest_src_size d size in
                                let s2 = make_dest_src_size s2 size in
                                let tac = {
                                        to_op = op;
                                        to_dest = d;
                                        to_src1 = s1;
                                        to_src2 = s2 } in
                                let gtac = (group, tac) in
                                let lgtac_out = gtac :: lgtac_out in
                                let lgtac_in = List.tl lgtac_in in
                                (tac_populate_group_sizes group size lgtac_in
                                        lgtac_out)
                        end
                else (* not interesting for check, but still size it *)
                        begin
                        let d = make_dest_src_size d size in
                        let s1 = make_dest_src_size s1 size in
                        let s2 = make_dest_src_size s2 size in
                        let tac = {
                                to_op = op;
                                to_dest = d;
                                to_src1 = s1;
                                to_src2 = s2 } in
                        let gtac = (group, tac) in
                        let lgtac_out = gtac :: lgtac_out in
                        let lgtac_in = List.tl lgtac_in in
                        (tac_populate_group_sizes group size lgtac_in
                                lgtac_out)
                        end
                end

(*
 * get_max_group
 *
 *      Get the maximum group number in the list of gtac elements.
 *
 * Output:  (max number)
 *)
let rec get_max_group max lgtac =
        if lgtac == [] then max 
        else
        let group = fst (List.hd lgtac) in
        let max = 
                (if group > max then group else max) in
        let lgtac = List.tl lgtac in
        (get_max_group max lgtac)

(*
 * tac_populate_sizes
 *
 *      Populate element sizes for all groups within the list of
 *      grouped TAC elements.
 *
 * Output:  (lgtac)
 *)
let rec tac_populate_sizes group max lgtac =
        if group > max then lgtac
        else
                begin
                let size = tac_get_expr_size_for_group group lgtac in
                let lgtac = tac_populate_group_sizes group size 
                        lgtac [] in
                let lgtac = List.rev lgtac in
                let group = group + 1 in
                (tac_populate_sizes group max lgtac)
                end

let tac_expr_sizing lgtac =
        let max = get_max_group 0 lgtac in
        let lgtac = tac_populate_sizes 0 max lgtac in
        let ltac = List.map snd lgtac in
        (ltac)


(****************************************************************
 *
 * Routines for Evaluating Parser ASTs and Creating TAC Lists
 *
 ****************************************************************)

(*
 * tac_eval_expr
 *
 *      Evaluate an expression provided by the parser, and from it
 *      append elements to a list of Three Address Code.  The expressions
 *      are evaluated, and returned is the updated environment, the 
 *      currend TAC destination where the result is stored (if a temporary
 *      result is required), and the appended TAC list.
 *
 * Output:  (env, TAC dest/src, ltac) 
 *)
let rec tac_eval_expr env gsymb lsymb expr_in tac_list =
        match expr_in with
            Literal(i) ->
                    let curr_dest = { 
                            tds_type = Dimm;
                            tds_name = "invld";
                            tds_value = i;
                            tds_size = 4 } in
                    (env, curr_dest, tac_list)
          | Id(n) ->
                    let curr_symb = 
                        (if NameMap.mem n gsymb then NameMap.find n gsymb
                         else if NameMap.mem n lsymb then NameMap.find n lsymb
                         else raise (Failure ("undeclared variable " ^ n))) in
                    if curr_symb.symb_stype != Svar then
                            raise (Failure ("non-variable used as variable: " ^ n))
                    else let curr_dest = {
                            tds_type = Dvar;
                            tds_name = n;
                            tds_value = 0;
                            tds_size = curr_symb.symb_size } in
                    (env, curr_dest, tac_list)
          | Array(n, e, dt) ->
                    (* evaluate the index expression *)
                    let ind_env = {
                            te_label = env.te_label;
                            te_trslt = env.te_trslt;
                            te_seed = env.te_seed;
                            te_group = env.te_group + 1 } in
                    let temp_group = env.te_group in
                    let (env, ind_dest, tac_list) = 
                            tac_eval_expr ind_env gsymb lsymb e tac_list in
                    (* restore the current group *)
                    let env = {
                            te_label = env.te_label;
                            te_trslt = env.te_trslt;
                            te_seed = env.te_seed;
                            te_group = temp_group } in
                    let curr_symb = 
                        (if NameMap.mem n gsymb then NameMap.find n gsymb
                         else if NameMap.mem n lsymb then NameMap.find n lsymb
                         else raise (Failure ("undeclared variable " ^ n))) in
                    if curr_symb.symb_stype != Sarray then
                            raise (Failure ("non-array used as array: " ^ n))
                    else
                        let ind_size = 
                                curr_symb.symb_size / (get_size_from_data_type
                                        curr_symb.symb_dtype) in
                        let ind_size_src = {
                                tds_type = Dimm;
                                tds_name = "invld";
                                tds_value = ind_size;
                                tds_size = 0 } in
                        (* get a trslt for the index, and place it in there *)
                        let (ind_trslt, env) = get_trslt_dest env 4 in
                        (* < asn, dest, src, invld > *)
                        let tac_list = append_tac_list_g (env.te_group+1) Tasn
                                ind_trslt ind_dest dest_src_invld tac_list in
                        (* < rdiv, ind_dest, ind_dest, ind_size_src > *)
                        let tac_list = append_tac_list_g (env.te_group+1) Trdiv
                                ind_trslt ind_trslt ind_size_src tac_list in
                        let dsz = get_size_from_data_type dt in
                        let (trslt, env) = get_trslt_dest env dsz in
                        let array_dest = {
                                tds_type = Darray;
                                tds_name = n;
                                tds_value = dsz;
                                tds_size = curr_symb.symb_size } in
                        (* < array, trslt, array, index > *)
                        let tac_list = append_tac_list_g env.te_group Tarray
                                trslt array_dest ind_trslt tac_list in
                        (env, trslt, tac_list)
          | Binop(e1, op, e2) ->
                let (env, src1, tac_list) = tac_eval_expr env gsymb lsymb e1
                        tac_list in
                let (env, src2, tac_list) = tac_eval_expr env gsymb lsymb e2
                        tac_list in
                let (trslt, env) = get_trslt_dest env 4 in
                (* < op, trslt, src1, src2 > *)
                let tacop = get_tac_op_from_binop op in
                let tac_list = append_tac_list_g env.te_group tacop trslt src1 src2 tac_list in
                (env, trslt, tac_list)
          | Uniop(op, e) ->
                let (env, src1, tac_list) = tac_eval_expr env gsymb lsymb e
                        tac_list in
                let (trslt, env) = get_trslt_dest env 4 in
                let tacop = get_tac_op_from_uniop op in
                (* < op, trslt, src1, invld > *)
                let tac_list = append_tac_list_g env.te_group tacop trslt src1 dest_src_invld
                        tac_list in
                (env, trslt, tac_list)
          | Assign(n, e) ->
                let (env, src1, tac_list) = tac_eval_expr env gsymb lsymb e
                        tac_list in
                let curr_symb = 
                        (if NameMap.mem n gsymb then NameMap.find n gsymb
                         else if NameMap.mem n lsymb then NameMap.find n lsymb
                         else raise (Failure ("undeclared variable " ^ n))) in
                if curr_symb.symb_stype != Svar then
                        raise (Failure ("non-variable used as variable: " ^ n))
                else let curr_dest = {
                        tds_type = Dvar;
                        tds_name = n;
                        tds_value = 0;
                        tds_size = curr_symb.symb_size } in
                (* < Tasn, dest, src1, invld > *)
                let tac_list = append_tac_list_g env.te_group Tasn curr_dest src1
                        dest_src_invld tac_list in
                (env, curr_dest, tac_list)
          | Arrayassign(n, e1, dt, e2) ->
                let (env, src2, tac_list) = tac_eval_expr env gsymb lsymb e2
                        tac_list in
                (* index source is in a seperate group for sizings *)
                let temp_group = env.te_group in
                let env = {
                        te_label = env.te_label;
                        te_trslt = env.te_trslt;
                        te_seed = env.te_seed;
                        te_group = env.te_group + 1 } in
                let (env, ind_src, tac_list) = tac_eval_expr env gsymb lsymb e1
                        tac_list in
                (* restore the current group *)
                let env = {
                        te_label = env.te_label;
                        te_trslt = env.te_trslt;
                        te_seed = env.te_seed;
                        te_group = temp_group } in
                let curr_symb = 
                        (if NameMap.mem n gsymb then NameMap.find n gsymb
                         else if NameMap.mem n lsymb then NameMap.find n lsymb
                         else raise (Failure ("undeclared variable " ^ n))) in
                if curr_symb.symb_stype != Sarray then
                        raise (Failure ("non-array used as array: " ^ n))
                else 
                        let ind_size = 
                                curr_symb.symb_size / (get_size_from_data_type
                                        curr_symb.symb_dtype) in
                        let ind_size_src = {
                                tds_type = Dimm;
                                tds_name = "invld";
                                tds_value = ind_size;
                                tds_size = 0 } in
                        (* get a trslt for the index, and place it in there *)
                        let (ind_trslt, env) = get_trslt_dest env 4 in
                        (* < asn, dest, src, invld > *)
                        let tac_list = append_tac_list_g (env.te_group+1) Tasn
                                ind_trslt ind_src dest_src_invld tac_list in
                        (* < rdiv, ind_src, ind_src, ind_size_src > *)
                        let tac_list = append_tac_list_g (env.te_group+1) Trdiv
                                ind_trslt ind_trslt ind_size_src tac_list in
                        let (trslt, env) = get_trslt_dest env 4 in
                        let array_dest = {
                                tds_type = Darray;
                                tds_name = n;
                                tds_value = (get_size_from_data_type dt);
                                tds_size = curr_symb.symb_size } in
                        (* < aasn, array, ind_reg, src2 > *)
                        let tac_list = append_tac_list_g env.te_group Taasn array_dest
                                ind_trslt src2 tac_list in
                        (* < asn, trslt, src2, invld > *)
                        let tac_list = append_tac_list_g env.te_group Tasn trslt src2
                                dest_src_invld tac_list in
                        (env, trslt, tac_list)
          | Call(n, el) ->
                let el = List.rev el in
                let fsymb = 
                        (if NameMap.mem n gsymb then NameMap.find n gsymb
                         else raise (Failure ("undeclared function " ^ n))) in
                if fsymb.symb_stype != Sfunc then
                         raise (Failure ("non-variable used as variable: " ^ n))
                else
                        begin
                        let rec push_params env el psl tac_list =
                                if el == [] then
                                      begin
                                      if psl != [] then
                                            raise (Failure ("Function called
                                            without correct number of arguments:
                                            " ^ n))
                                      else (env, dest_src_invld, tac_list)
                                      end
                                else
                                      begin
                                      if psl == [] then
                                            raise (Failure ("Function called
                                            with incorrect number of arguments: "
                                            ^ n))
                                      else
                                            begin
                                            let ce = List.hd el in
                                            let el = List.tl el in
                                            let ps = List.hd psl in
                                            let psl = List.tl psl in
                                            let (env, trslt, tac_list) =
                                                  tac_eval_expr env gsymb lsymb
                                                        ce tac_list in
                                            let trslt = {
                                                          tds_type = trslt.tds_type;
                                                          tds_name = trslt.tds_name;
                                                          tds_value = trslt.tds_value;
                                                          tds_size = ps } in
                                            (* < Tparam, src_param, invld, invld > *)
                                            let tac_list = append_tac_list_g env.te_group
                                                  Tparam trslt dest_src_invld 
                                                  dest_src_invld tac_list in
                                            push_params env el psl tac_list
                                            end
                                      end
                        in
                        let (env, _, tac_list) = push_params env el
                                fsymb.symb_paramsizes tac_list
                        in
                        let num_params = List.length fsymb.symb_paramsizes in
                        let sz = get_size_from_data_type fsymb.symb_dtype in
                        let (rdest, env) = get_trslt_dest env sz in
                        let callsrc = {
                                tds_type = Dvar;
                                tds_name = n;
                                tds_value = 0;
                                tds_size = sz } in
                        (* finally, we make the call *)
                        (* < call, return_dest, func name, invld > *)
                        let tac_list = append_tac_list_g env.te_group Tcall rdest callsrc 
                                dest_src_invld tac_list in
                        (* following return, we must adjust the stack *)
                        (* add 4 for every parameter *)
                        let espsrc = {
                                tds_type = Dimm;
                                tds_name = "invld";
                                tds_value = (4 * num_params);
                                tds_size = 0
                                } in
                        (* < Tespadd, value, invld, invld > *)
                        let tac_list = append_tac_list_g env.te_group Tespadd espsrc
                                dest_src_invld dest_src_invld tac_list in
                        (env, rdest, tac_list)
                        end
          | Paren(e) ->
                let (env, trslt, tac_list) = tac_eval_expr env gsymb lsymb e
                        tac_list in
                (env, trslt, tac_list)
          | Noexpr -> (env, dest_src_invld, tac_list)
                


let rec append_lists l1 l2 =
        (* note: list 2 should be reversed *)
        if l2 == [] then l1
        else 
                let l1 = (List.hd l2) :: l1 in
                (append_lists l1 (List.tl l2))        

(*
 * tac_eval_stmt
 *
 *      Evaluate a statement provided by the parser in the form of an AST
 *      and from it generate a list of TAC elements representing the 
 *      statements execution.
 *
 *      Note here that for consistency, these statements return a TAC
 *      dest/src, just like the expression evaluation does.  This is 
 *      not used, however.
 *
 * Output:  (env, TAC dest/src, ltac)
 *)
let rec tac_eval_stmt env gsymb lsymb stmt_in tac_list =
        match stmt_in with
            Block(sl) -> 
                    if sl == [] then (env, dest_src_invld, tac_list)
                    else
                        let (env, _, tac_list) =
                            tac_eval_stmt env gsymb lsymb (List.hd sl) tac_list
                        in
                        let (env, _, tac_list) =
                            tac_eval_stmt env gsymb lsymb (Block(List.tl sl))
                                tac_list in
                        (env, dest_src_invld, tac_list)
          | Expr(e) ->
                    let (env, _, gtac_list) = 
                            tac_eval_expr env gsymb lsymb e [] in
                    let tmp_tac_list = tac_expr_sizing gtac_list in
                    let tac_list = append_lists tac_list (List.rev tmp_tac_list)
                    in
                    (env, dest_src_invld, tac_list)
          | Return(e) ->
                    let (env, dest, gtac_list) =
                            tac_eval_expr env gsymb lsymb e [] in
                    if dest.tds_type == Darray then
                           raise (Failure ("Return type array " ^
                                dest.tds_name))
                    else 
                    (* < label, label_dest, invld, invld > *)
                    let gtac_list = append_tac_list_g env.te_group Tret dest
                            dest_src_invld dest_src_invld gtac_list in
                    let tmp_tac_list = tac_expr_sizing gtac_list in
                    let tac_list = append_lists tac_list (List.rev tmp_tac_list)
                    in
                    (env, dest_src_invld, tac_list)
          | If(e, s1, s2) ->
                    if s2 == Nostmt then
                        let (env, de, gtac_list) = 
                                tac_eval_expr env gsymb lsymb e [] in
                        (* evict all registers *)
                        let gtac_list = append_tac_list_g env.te_group Tevictall
                                dest_src_invld dest_src_invld  dest_src_invld gtac_list in
                        (* < cmp, invld, de, zero > *)
                        let gtac_list = append_tac_list_g env.te_group Tcmp dest_src_invld de
                              dest_src_zero gtac_list in 
                        let (jlbl, env) = get_label_dest env in
                        (* < jmpe, L$, invld, invld > *)
                        let gtac_list = append_tac_list_g env.te_group Tjmpe jlbl 
                              dest_src_invld dest_src_invld gtac_list in
                        let tmp_tac_list = tac_expr_sizing gtac_list in
                        let tac_list = append_lists tac_list (List.rev tmp_tac_list)
                        in
                        let (env, _, tac_list) = 
                              tac_eval_stmt env gsymb lsymb s1 tac_list in
                        (* evict all registers before leaving loop body *)
                        let tac_list = append_tac_list Tevictall dest_src_invld
                                dest_src_invld dest_src_invld tac_list in
                        (* < label, L$, invld, invld > *)
                        let tac_list = append_tac_list Tlabel jlbl
                              dest_src_invld dest_src_invld tac_list in
                        (env, dest_src_invld, tac_list)
                    else
                        let (env, de, gtac_list) = 
                                tac_eval_expr env gsymb lsymb e [] in
                        (* evict all registers before entering loop body *)
                        let gtac_list = append_tac_list_g env.te_group Tevictall
                                dest_src_invld dest_src_invld dest_src_invld
                                gtac_list in
                        (* < cmp, invld, de, zero > *)
                        let gtac_list = append_tac_list_g env.te_group Tcmp dest_src_invld de
                              dest_src_zero gtac_list in
                        let tmp_tac_list = tac_expr_sizing gtac_list in
                        let tac_list = append_lists tac_list (List.rev tmp_tac_list)
                        in 
                        let (jlbl, env) = get_label_dest env in
                        (* < jmpe, L$, invld, invld > *)
                        let tac_list = append_tac_list Tjmpe jlbl 
                              dest_src_invld dest_src_invld tac_list in
                        let (env, _, tac_list) = 
                              tac_eval_stmt env gsymb lsymb s1 tac_list in
                        let (jlbl2, env) = get_label_dest env in
                        (* evict all registers before leaving this if block *)
                        let tac_list = append_tac_list Tevictall dest_src_invld
                                dest_src_invld dest_src_invld tac_list in
                        (* < jmp, jlbl2, invld, invld > *)
                        let tac_list = append_tac_list Tjmp jlbl2
                              dest_src_invld dest_src_invld tac_list in
                        (* < label, L$, invld, invld > *)
                        let tac_list = append_tac_list Tlabel jlbl
                              dest_src_invld dest_src_invld tac_list in
                        let (env, _, tac_list) = 
                              tac_eval_stmt env gsymb lsymb s2 tac_list in
                        (* evict all registers before leaving this if block *)
                        let tac_list = append_tac_list Tevictall dest_src_invld
                                dest_src_invld dest_src_invld tac_list in
                        (* < lable, jlbl2, invld, invld > *)
                        let tac_list = append_tac_list Tlabel jlbl2 
                              dest_src_invld dest_src_invld tac_list in
                        (env, dest_src_invld, tac_list)
          | For(e1, e2, e3, s) ->
                        (* start evaluation of expressions *)
                        let (env, _, gtac_list) = tac_eval_expr env gsymb lsymb
                                e1 [] in
                        (* Evict all registers before starting loop *)
                        (* < Tevictall, invld, invld, invld > *)
                        let gtac_list = append_tac_list_g env.te_group Tevictall dest_src_invld
                                dest_src_invld dest_src_invld gtac_list in
                        let tmp_tac_list = tac_expr_sizing gtac_list in
                        let tac_list = append_lists tac_list (List.rev tmp_tac_list)
                        in
                        let (jlbl, env) = get_label_dest env in
                        (* < label, jlbl, invld, invld > *)
                        let tac_list = append_tac_list Tlabel jlbl
                                dest_src_invld dest_src_invld tac_list in
                        let (env, trslt, gtac_list) = tac_eval_expr env gsymb
                                lsymb e2 [] in
                        (* < cmp, invld, trslt, zero > *)
                        let gtac_list = append_tac_list_g env.te_group Tcmp dest_src_invld
                                trslt dest_src_zero gtac_list in
                        let tmp_tac_list = tac_expr_sizing gtac_list in
                        let tac_list = append_lists tac_list (List.rev tmp_tac_list)
                        in
                        let (jlbl2, env) = get_label_dest env in
                        (* < jmpe, jlbl2, invld, invld > *)
                        let tac_list = append_tac_list Tjmpe jlbl2 
                                dest_src_invld dest_src_invld tac_list in
                        let (env, _, tac_list) = tac_eval_stmt env gsymb lsymb
                                s tac_list in
                        let (env, trslt2, gtac_list) = tac_eval_expr env gsymb
                                lsymb e3 [] in
                        let tmp_tac_list = tac_expr_sizing gtac_list in
                        let tac_list = append_lists tac_list (List.rev tmp_tac_list)
                        in
                        (* Evict before looping back *)
                        (* < Tevictall, invld, invld, invld > *)
                        let tac_list = append_tac_list Tevictall dest_src_invld
                                dest_src_invld dest_src_invld tac_list in
                        (* < jmp, jlbl, invld, invld > *)
                        let tac_list = append_tac_list Tjmp jlbl dest_src_invld
                                dest_src_invld tac_list in
                        (* < label, jlbl2, invld, invld > *)
                        let tac_list = append_tac_list Tlabel jlbl2 
                                dest_src_invld dest_src_invld tac_list in
                        (env, dest_src_invld, tac_list)
          | While(e, s) ->
                        (* Evict all registers before starting loop *)
                        (* < Tevictall, invld, invld, invld > *)
                        let tac_list = append_tac_list Tevictall dest_src_invld
                                dest_src_invld dest_src_invld tac_list in
                        let (jlbl1, env) = get_label_dest env in
                        (* < label, jlbl1, invld, invld > *)
                        let tac_list = append_tac_list Tlabel jlbl1 
                                dest_src_invld dest_src_invld tac_list in
                        let (env, trslt, gtac_list) = 
                                tac_eval_expr env gsymb lsymb e [] in
                        (* < cmp, invld, trslt, zero > *)
                        let gtac_list = append_tac_list_g env.te_group Tcmp dest_src_invld trslt 
                                dest_src_zero gtac_list in
                        let tmp_tac_list = tac_expr_sizing gtac_list in
                        let tac_list = append_lists tac_list (List.rev tmp_tac_list)
                        in
                        let (jlbl2, env) = get_label_dest env in
                        (* < jmpe, jlbl2, invld, invld > *)
                        let tac_list = append_tac_list Tjmpe jlbl2
                                dest_src_invld dest_src_invld tac_list in
                        let (env, _, tac_list) =
                                tac_eval_stmt env gsymb lsymb s tac_list in
                        (* Evict all registers before looping *)
                        (* < Tevictall, invld, invld, invld > *)
                        let tac_list = append_tac_list Tevictall dest_src_invld
                                dest_src_invld dest_src_invld tac_list in
                        (*
                         * We must evict again before jumping back - this
                         * guarentees that between loop iterations the assembly
                         * is in the same state. 
                         *)
                        (* < Tevictall, invld, invld, invld > *)
                        let tac_list = append_tac_list Tevictall dest_src_invld
                                dest_src_invld dest_src_invld tac_list in
                        (* < jmp, jlbl1, invld, invld > *)
                        let tac_list = append_tac_list Tjmp jlbl1 dest_src_invld
                                dest_src_invld tac_list in
                        (* < label, jlbl2, invld, invld > *)
                        let tac_list = append_tac_list Tlabel jlbl2 
                                dest_src_invld dest_src_invld tac_list in
                        (env, dest_src_invld, tac_list)
          | Rcbcall(rl) ->
                        let (env, seed) = get_seed env in
                        let rec determine_rcb seed rl sum =
                                if rl == [] then (0, "invld")
                                else let r = List.hd rl in
                                if (sum + (fst r) > seed) then (1, (snd r))
                                else determine_rcb seed (List.tl rl)
                                        (sum + (fst r))
                        in
                        let (use, n) = determine_rcb seed rl 0 in
                        (* if we are not calling an RCB, return *)
                        if use == 0 then (env, dest_src_invld, tac_list)
                        else let rsymb = 
                                (if NameMap.mem n gsymb then NameMap.find n gsymb
                                else raise (Failure ("undeclared rcb " ^ n))) in
                        let callsrc = {
                                tds_type = Dvar;
                                tds_name = n;
                                tds_value = 0;
                                tds_size = 0 } in
                        (* this becomes just a call to the RCB *)
                        (* 
                         * even an RCB must specify a tmp return - even though
                         * it wont use it.
                         *)
                        let (rdest, env) = get_trslt_dest env 4 in
                        (* < call, return_dest, call_src, invld > *)
                        let tac_list = append_tac_list Tcall rdest  
                                callsrc dest_src_invld tac_list in
                        (env, dest_src_invld, tac_list)
          | Nostmt -> (env, dest_src_invld, tac_list)


let rec tac_eval_stmt_list env gsymb lsymb sl tac_list =
        if sl == [] then (env, dest_src_invld, tac_list)
        else
                let s = List.hd sl in
                let (env, _, tac_list) = 
                        tac_eval_stmt env gsymb lsymb s
                        tac_list in
                let (env, _, tac_list) = 
                        tac_eval_stmt_list env gsymb lsymb
                                (List.tl sl) tac_list in
                (env, dest_src_invld, tac_list)
                                

(****************************************************************
 *
 * Routines for Generating TAC Elements at the Function+ Level
 *
 ****************************************************************)

(*
 * create_func
 *
 *      Create a function using the function type defined in
 *      tac_types.ml.  This will populate a function with 
 *      the function symbol, globals, locals, and a list of
 *      TAC elements representing the function.
 *
 * Output:  (env, function type)
 *)                        
let create_func env gsymb fin =
        let n = fin.fdef_name in
        let fsymb = 
                (if NameMap.mem n gsymb then NameMap.find n gsymb
                 else raise (Failure ("undeclared function " ^ n))) in
        let lsymb = NameMap.empty in
        let lsymb = populate_symbols fin.fdef_vdecls [] [] fin.fdef_params lsymb
        in
        let (env, _, tac_list) = tac_eval_stmt_list env gsymb lsymb
                                        fin.fdef_body [] in
        let tac_list = List.rev tac_list in
        let func_out = {
                func_symb = fsymb;
                func_globals = gsymb;
                func_locals = lsymb;
                func_ltac = tac_list } in
        (env, func_out)


(*
 * create_rcb
 *
 *      Create an RCB from the AST provided by the parser.
 *
 * Output:  (env, RCB type)
 *)
let create_rcb env gsymb rin =
        let n = rin.rcbdef_name in
        let rsymb = 
                (if NameMap.mem n gsymb then NameMap.find n gsymb
                else raise (Failure ("undeclared rcb " ^ n))) in
        let lsymb = NameMap.empty in
        let lsymb = populate_symbols rin.rcbdef_vdecls [] [] [] lsymb in
        let (env, _, tac_list) = tac_eval_stmt_list env gsymb lsymb
                                        rin.rcbdef_body [] in
        let tac_list = List.rev tac_list in
        let rcb_out = {
                rcb_symb = rsymb;
                rcb_globals = gsymb;
                rcb_locals = lsymb;
                rcb_ltac = tac_list } in
        (env, rcb_out)


(*
 * tac_create_prog
 *      
 *      Generate an entire program from the tupple generated by the
 *      parser.  This routine is the top-level call which actually
 *      evaluates and provides Three Address Code representation for
 *      all elements.
 *
 * Output:  (program)
 *)
let one_of_six = (fun (x, _, _, _, _, _) -> x)
let two_of_six = (fun (_, x, _, _, _, _) -> x)
let three_of_six = (fun (_, _, x, _, _, _) -> x)
let four_of_six = (fun (_, _, _, x, _, _) -> x)
let five_of_six = (fun (_, _, _, _, x, _) -> x)
let six_of_six = (fun (_, _, _, _, _, x) -> x)

let tac_create_prog pin seed =
        let env = {
                te_label = 0;
                te_trslt = 0;
                te_seed = seed;
                te_group = 0
                } in
        let gsymb = NameMap.empty in
        let gsymb = populate_symbols (one_of_six pin) (two_of_six pin)
                        (three_of_six pin) [] gsymb in
        let rec create_rcbs env rcbl_in rcbl_out =
                if rcbl_in == [] then (env, rcbl_out)
                else let (env, curr_rcb) = create_rcb env gsymb 
                                                (List.hd rcbl_in) in
                create_rcbs env (List.tl rcbl_in) (curr_rcb :: rcbl_out)
        in
        let (env, rcbs) = create_rcbs env (six_of_six pin) [] in
        let rec create_funcs env fl_in fl_out =
                if fl_in == [] then (env, fl_out)
                else let (env, curr_f) = create_func env gsymb
                                                (List.hd fl_in) in
                create_funcs env (List.tl fl_in) (curr_f :: fl_out)
        in
        let (env, funcs) = create_funcs env (five_of_six pin) [] in
        let prog_out = {
                prog_globals = gsymb;
                prog_lrcb = rcbs;
                prog_lfunc = funcs } in
        (prog_out)








                            



